/** @preserve Eloquence WEBDLG v2
 (C) Copyright Marxmeier Software AG, 2020-2021
 Version: 1.954 2021-07-16
 $Id: 00-base.js,v 28.44 2021/07/16 12:46:24 mike Exp $
*//*======================================================================
   WEBDLG v2 namespace / globals
========================================================================*/

var eq = {
   c : {},           // Classes
   p : {             // Properties:
      version : 1954 //  Version 1.954 (major*1,000 + minor)
   }
};

/*========================================================================
   class Obj
   Generic base class.
========================================================================*/

eq.c.Obj = function() {};

/*------------------------------------------------------------------------
   static Obj.subClass()
   Subclassing convenience helper.
------------------------------------------------------------------------*/

(eq.c.Obj.setClassName = function(name, cls) {
   var pd = Object.getOwnPropertyDescriptor(cls.prototype.constructor, 'name');
   if (pd !== undefined)
      pd.value = name;
   else
      pd = {
         value : name,
         configurable : true,
         enumerable : false,
         writable : false
      };
   Object.defineProperty(cls.prototype.constructor, 'name', pd);
})("Obj", eq.c.Obj);

eq.c.Obj.subClass = function(name, cls) {
   var base = this;
   cls.prototype = Object.create(base.prototype);
   Object.defineProperty(cls.prototype, 'constructor', {
      value : cls,
      configurable : false,
      enumerable : false,
      writable : true
   });
   eq.c.Obj.setClassName(name, cls);
   cls.subClass = function(n, c) { return base.subClass.call(this, n, c) };
   return eq.c[name] = cls;
};

/** @preserve $Id: 05-comm.js,v 28.19 2021/07/13 13:16:06 mike Exp $
*//*======================================================================
   class Rq extends Obj
   DLG request.
========================================================================*/

eq.c.Obj.subClass('Rq', function() {
   eq.c.Obj.call(this);
});

/*------------------------------------------------------------------------
   Rq.dispose()
   Dispose request data buffer.
------------------------------------------------------------------------*/

eq.c.Rq.prototype.dispose = function() {
   this.bf = undefined;
   this.bfOffset = undefined;
   this.bfLength = undefined;
};

/*------------------------------------------------------------------------
   Rq.assign()
   Assign request data buffer.
------------------------------------------------------------------------*/

eq.c.Rq.prototype.assign = function(buffer) {
   this.bf = new Uint8Array(buffer);
   this.bfOffset = 0;
   this.bfLength = this.bf.length;
};

/*------------------------------------------------------------------------
   Rq.decodeUi32()
   Decode varint-encoded 32-bit unsigned integer.
------------------------------------------------------------------------*/

eq.c.Rq.prototype.decodeUi32 = function() {
   var iv = 0, shift = 0, b;
   for (var cnt = 0;;) {
      if (this.bfOffset >= this.bfLength)
         throw new Error("DLG buffer offset exceeded: " +
                         (this.bfOffset + 1) + ", max: " + this.bfLength);
      if (++cnt > 5)
         throw new Error("DLG buffer varint overflow, offset: " +
                         this.bfOffset);

      b = this.bf[this.bfOffset++];
      if (!(b & 0x80))
         break;

      iv |= (b & 0x7f) << shift;
      shift += 7;
   }
   return iv | b << shift;
};

/*------------------------------------------------------------------------
   Rq.decodeUint()
   Decode varint-encoded unsigned integer.
------------------------------------------------------------------------*/

eq.c.Rq.prototype.decodeUint = function() {
   var off = this.bfOffset, iv = this.decodeUi32();
   if (iv < 0)
      throw new Error("DLG buffer UINT out of range: " +
                      iv + ", offset: " + off);
   return iv;
};

/*------------------------------------------------------------------------
   Rq.decodeInt()
   Decode varint-encoded signed integer.
------------------------------------------------------------------------*/

eq.c.Rq.prototype.decodeInt = function() {
   var iv = this.decodeUi32();
   return iv & 1 ? iv >>> 1 ^ -1 : iv >>> 1;
};

/*------------------------------------------------------------------------
   Rq.decodeString()
   Decode varint-encoded string.
------------------------------------------------------------------------*/

eq.c.Rq.decodeStringFailed = function(off, len, minLen) {
   throw new Error("DLG buffer invalid STRING UTF-8 encoding, offset: " +
                   off + ", len: " + len + ", min: " + minLen);
};

eq.c.Rq.prototype.decodeString = function() {
   var len = this.decodeUint(), off = this.bfOffset;

   this.bfOffset += len;
   if (this.bfOffset > this.bfLength)
      throw new Error("DLG buffer offset exceeded: " +
                      this.bfOffset + ", max: " + this.bfLength);

   var s = "";
   while (len) {
      var b = this.bf[off];

      if (b < 0x80) {
         s += String.fromCharCode(b);
         off += 1;
         len -= 1;
      }
      else if (b > 0xbf && b < 0xe0) {
         if (len < 2)
            eq.c.Rq.decodeStringFailed(off, len, 2);

         s += String.fromCharCode(  (b & 0x1f) << 6
                                  | this.bf[off+1] & 0x3f);
         off += 2;
         len -= 2;
      }
      else if (b > 0xdf && b < 0xf0) {
         if (len < 3)
            eq.c.Rq.decodeStringFailed(off, len, 3);

         s += String.fromCharCode(  (b & 0xf) << 12
                                  | (this.bf[off+1] & 0x3f) << 6
                                  | this.bf[off+2] & 0x3f);
         off += 3;
         len -= 3;
      }
      else {
         // Surrogate pair.
         if (len < 4)
            eq.c.Rq.decodeStringFailed(off, len, 4);

         var c = (  (b & 0x7) << 18
                  | (this.bf[off+1] & 0x3f) << 12
                  | (this.bf[off+2] & 0x3f) << 6
                  | this.bf[off+3] & 0x3f) - 0x10000;
         s += String.fromCharCode(c >> 10 | 0xd800, c & 0x3ff | 0xdc00);
         off += 4;
         len -= 4;
      }
   }

   return s;
};

/*========================================================================
   class DlgRq extends Rq
   DLG request.
========================================================================*/

eq.c.Rq.subClass('DlgRq', function() {
   eq.c.Rq.call(this);
});

/*------------------------------------------------------------------------
   static DlgRq.mode
   DLG request modes.
------------------------------------------------------------------------*/

eq.c.DlgRq.mode = {
   ssnNew     : 1,  // New session
   ssnResume  : 2,  // Resume session
   global     : 3,  // Global properties
   add        : 4,  // Add control(s)
   modify     : 5,  // Modify control
   del        : 6,  // Delete control
   invalidate : 7,  // Invalidate dialog(s)
   DLG_DRAW   : 8,  // DLG DRAW
   DLG_DO     : 9,  // DLG DO
   POPUP_BOX  : 10  // POPUP BOX
};

/*------------------------------------------------------------------------
   DlgRq.dispose()
   Dispose request message.
------------------------------------------------------------------------*/

eq.c.DlgRq.prototype.dispose = function() {
   eq.c.Rq.prototype.dispose.call(this);
   this.msg = undefined;
};

/*------------------------------------------------------------------------
   Rq.assign()
   Assign request message.
------------------------------------------------------------------------*/

eq.c.DlgRq.prototype.assign = function(buffer) {
   eq.c.Rq.prototype.assign.call(this, buffer);
   this.msg = [];
};

/*========================================================================
   class Rs extends Obj
   DLG response.
========================================================================*/

eq.c.Obj.subClass('Rs', function() {
   eq.c.Obj.call(this);
   this.bf = [];
});

/*------------------------------------------------------------------------
   Rs.dispose()
   Dispose response data buffer.
------------------------------------------------------------------------*/

eq.c.Rs.prototype.dispose = function() {
   this.bf.length = 0;
};

/*------------------------------------------------------------------------
   Rs.buffer()
   Return and dispose response data buffer.
------------------------------------------------------------------------*/

eq.c.Rs.prototype.buffer = function() {
   var a = new Uint8Array(this.bf);
   this.bf.length = 0;
   return a.buffer;
};

/*------------------------------------------------------------------------
   Rs.encodeUint()
   Varint-encode unsigned integer.
------------------------------------------------------------------------*/

eq.c.Rs.prototype.encodeUint = function(iv) {
   var cnt = 0;
   while (iv & ~0x7f) {
      if (++cnt === 5)
         throw new Error("Varint value overflow");
      this.bf.push((iv & 0x7f) | 0x80);
      iv >>>= 7;
   }
   this.bf.push(iv);
};

/*------------------------------------------------------------------------
   Rs.encodeInt()
   Varint-encode signed integer.
------------------------------------------------------------------------*/

eq.c.Rs.prototype.encodeInt = function(iv) {
   this.encodeUint(iv < 0 ? iv << 1 ^ -1 : iv << 1);
};

/*------------------------------------------------------------------------
   Rs.encodeString()
   Varint-encode string.
------------------------------------------------------------------------*/

eq.c.Rs.prototype.encodeString = function(s) {
   var i, c, l = 0, sl = s.length;
   for (i = 0; i < sl; i++) {
      c = s.charCodeAt(i);
      if (c < 0x80)
         l += 1;
      else if (c < 0x800)
         l += 2;
      else if (c < 0xd800 || c >= 0xe000)
         l += 3;
      else {
         // Surrogate pair.
         if (++i === sl)
            throw new Error("Incomplete STRING surrogate pair, offset: " + i);
         l += 4;
      }
   }

   this.encodeUint(l);
   for (i = 0; i < sl; i++) {
      c = s.charCodeAt(i);
      if (c < 0x80)
         this.bf.push(c);
      else if (c < 0x800) {
         this.bf.push(0xc0 | (c >> 6),
                      0x80 | (c & 0x3f));
      }
      else if (c < 0xd800 || c >= 0xe000) {
         this.bf.push(0xe0 | (c >> 12),
                      0x80 | ((c >> 6) & 0x3f),
                      0x80 | (c & 0x3f));
      }
      else {
         // Surrogate pair.
         c = 0x10000
           + (((c & 0x3ff) << 10) | (s.charCodeAt(++i) & 0x3ff));
         this.bf.push(0xf0 | (c >>18),
                      0x80 | ((c >> 12) & 0x3f),
                      0x80 | ((c >> 6) & 0x3f),
                      0x80 | (c & 0x3f));
      }
   }
};

/*========================================================================
   class DlgRs extends Rs
   DLG response.
========================================================================*/

eq.c.Rs.subClass('DlgRs', function() {
   eq.c.Rs.call(this);
});

/*------------------------------------------------------------------------
   static DlgRs.op
   DLG response opcodes.
------------------------------------------------------------------------*/

eq.c.DlgRs.op = {
   start      : 1, // Initialize DLG session, start application
   resume     : 2, // Resume application
   draw       : 3, // DLG DRAW response
   event      : 4, // DLG DO response, arg: element id, event type
   popup      : 5, // POPUP BOX response, arg: button number
   invalidate : 6  // Invalidate dialog, arg: dialog id
};

/** @preserve $Id: 05-const.js,v 28.31 2021/07/16 16:19:43 mike Exp $
*//*======================================================================
   static MainOp
   API->Main message opcodes
========================================================================*/

eq.MainOp = {
   applicationStopped : 1,  // Data: restartable true/false
   applicationFailed  : 2,  // Data: none
   cancelCurrentCall  : 3,  // Data: none
   resume             : 4,  // Data: none
   global             : 5,  // Data: global property tag, value
   addControl         : 6,  // Data: dialog id, DOM element template
   modifyControl      : 7,  // Data: dialog id, DOM element template
   deleteControl      : 8,  // Data: dialog id, element id
   invalidate         : 9,  // Data: dialog id array
   dlg_DRAW           : 10, // Data: dialog id
   dlg_DO             : 11, // Data: dialog id, focus element id
   dlg_POPUP_BOX      : 12, // Data: POPUP BOX descriptor
   tabOrderUpdated    : 13  // Data: tab order array
};

/*========================================================================
   static ApiOp
   Main->API message opcodes
========================================================================*/

eq.ApiOp = {
   documentComplete : 1, // Data: layout grid
   openWebSocket    : 2, // Data: WebSocket URL
   onEvent          : 3, // Data: element id, event type
   onPopupEvent     : 4, // Data: button number
   onDraw           : 5, // Data: none
   invalidate       : 6, // Data: dialog id
   updateTabOrder   : 7  // Data: dialog id
};

/*========================================================================
   static DlgRq.tag
   DLG request tags.
========================================================================*/

eq.RqTag  = {
   child : 1, // Child control

   // Global properties

   baseUrl     : 2, // Base url
   helpBaseUrl : 3, // Help base url

   // Common attributes.

   cssid       : 10,  // CSS id
   cssclsDel   : 11,  // Delete CSS class(es)
   cssclsAdd   : 12,  // Add CSS class(es)
   raster      : 13,  // Raster
   posSize     : 14,  // Position/size
   size        : 15,  // Size, ToolBar children
   border      : 16,  // Border
   rule        : 17,  // Rule
   ruleMode    : 18, // When to submit rule (Tree/ListBox.singleclick)
   tabonenter  : 19, // Enter key navigates focus
   visible     : 20, // Visibility
   sensitive   : 21, // Sensitivity
   inactive    : 22, // Inactive, eg. TabBox tab
   tabStop     : 23, // Tab stop
   bgColor     : 24, // Background color
   fgColor     : 25, // Foreground color
   url         : 26, // Url

   // Dialog attributes.

   dialogPosSize : 30, // Dialog position/size
   fKey          : 31, // Function keys

   // GroupBox attributes.

   vWidth  : 40, // Visible width
   vHeight : 41, // Visible height

   // Button, Menu, StaticText attributes.

   text       : 50, // Text
   textPos    : 51, // Text position
   align      : 52, // Align
   active     : 53, // Active
   radioGroup : 54, // RadioButton group
   icon       : 55, // PushButton icon
   iconActive : 56, // PushButton icon when active

   // EditText attributes.

   multiline  : 60, // Multiline
   readonly   : 61, // Read-only
   content    : 62, // Content
   ispassword : 63, // IsPassword

   // ListBox, PopText, Tree attributes.

   columns    : 70, // Columns
   activeLine : 71, // Active line
   topItem    : 72, // First visible line
   status     : 73, // Status

   // TabBox attributes.

   tabs        : 80, // Tabs
   selectedTab : 81, // Selected tab

   // Splitter, ProgressBar attributes.

   panes       : 90, // Setup splitter panes
   vertical    : 91, // Split vertical
   position    : 92, // Divider position
   quickExpand : 93, // Quick-expand controls

   // ComboBox attributes.

   items   : 100, // List items
   boxSize : 101, // Box width/height

   // ProgressBar attributes.

   value : 110  // Value, empty=indeterminate
};

/*========================================================================
   static RsValType
   Response changed element value type.
========================================================================*/

eq.RsValType = {
   idFocus       : 1,  // Focus object id, STRING
   ruleOnce      : 2,  // Submit this rule once, INT
   textValue     : 3,  // Text value, STRING
   buttonChecked : 4,  // Button checked, UINT
   caretPos      : 5,  // Caret position, UINT
   activeLine    : 6,  // Active line, UINT
   topItem       : 7,  // First visible line, UINT
   selectedTab   : 8,  // Selected tab, UINT
   splitterPos   : 9,  // Splitter position, UINT
   status        : 10, // ListBox/Tree status, ARRAY
   ruleKey       : 11, // Rule submission key code, UINT
   mouseBtns     : 12, //                 mouse buttons, UINT
   mouseCnt      : 13, //                 mouse click count, UINT
   mouseMods     : 14, //                 mouse modifier keys, UINT
};

/** @preserve $Id: 05-dlg.js,v 28.30 2021/07/16 16:19:43 mike Exp $
*//*======================================================================
   WEBDLG v2 namespace / globals
   List of DLG control classes.
========================================================================*/

eq.d = [];

/*========================================================================
   class Dlg extends Obj
   DLG dialog instance.
========================================================================*/

eq.c.Obj.subClass('Dlg', function(id) {
   eq.c.Obj.call(this);
   this.id = id;        // Dialog id.
   this.root = null;    // Root control.
   this.tabOrder = [];  // Tab order.
   this.tabOrderValid = false;
});

/*------------------------------------------------------------------------
   Dlg.close()
   Close dialog instance.
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.close = function() {
   if (this.root) {
      this.root.del();
      this.root = null;
   }

   this.tabOrder.length = 0;
   this.tabOrderValid = false;
};

/*------------------------------------------------------------------------
   Dlg.addChild()
   Add control from DLG request.
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.addChild = function(rq, ctParent, etParent) {
   // Control class, create control, set id.
   var ct = new eq.d[rq.decodeUint() - 1]();
   ct.id = rq.decodeString();

   // Add to parent.
   if (ctParent) {
      ctParent.addChild(ct);
      if (ct.focusable && !ctParent.containerFocusable())
         ct.focusable = false;
   }
   else {
      if (this.root)
         this.root.del();
      this.root = ct;
   }

   // Setup control, create children if any.

   var et = {
      // DOM element template.
      // See: eq.c.App.prototype.addControls()

      id : ct.id,              // Element id
      cb : ct.constructor.name // Callback, DLG class name

      // pa  : "" Parent element id
      // tg  : "" Tag
      // cla : [] Add class names
      // clr : [] Remove class names
      // at  : {} Attributes
      // st  : {} Styles
      // hc  : "" Text or HTML content
      // hca : "" Text or HTML content w/ optional accelerator
      // ec  : "" EditText content
      // rt  : [] Value by request tag
      // ev  : {  Event(s):
      //        a : [] Action event type(s)
      //        c : [] Change event type(s)
      //        m : Expected number of mouse clicks
      //        f : Callback-specific event flags
      //       }
      // dlg : {  Dialog properties:
      //        l : {  Layout
      //             x, y, w, h
      //             resizable
      //            }
      //        t : "" Title
      //        f : [] Function key rule values
      //       }
      // et  : {} Nested template
      // ch  : [] Children
   };

   if (etParent) {
      etParent.ch = etParent.ch || [];
      etParent.ch.push(et);
   }
   else {
      if (ctParent)
         et.pa = ctParent.id;
      rq.msg.push({
         o : eq.MainOp.addControl,
         i : this.id,
         d : et
      });
   }

   ct.setup(this, rq, et);

   if (ct.focusable && ct.visible && ct.sensitive) {
      var cl = et.cla = et.cla || [];
      cl.push('eq-focus');
   }
   else if (et.cla !== undefined) {
      var cl = et.cla, i = cl.indexOf('eq-noenter');
      if (i !== -1)
         cl.splice(i, 1);
   }
}

/*------------------------------------------------------------------------
   Dlg.findById()
   Find control by id.
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.findById = function(id) {
   var found = this.root ? this.root.findById(id) : null;
   if (found)
      return found;

   throw new Error("Dlg.findById(" + id + ") failed: not found");
}

/*------------------------------------------------------------------------
   Dlg.updateTabOrder()
   Fill current tab order array.
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.updateTabOrder = function() {
   this.tabOrder.length = 0;
   if (this.tabOrderValid = this.root !== null)
      this.root.updateTabOrder(this.tabOrder);
}

/*========================================================================
   class DlgControl extends Obj
   DLG control base class.
========================================================================*/

eq.c.Obj.subClass('DlgControl', function() {
   eq.c.Obj.call(this);
});

eq.c.DlgControl.prototype.posRaster = true;
eq.c.DlgControl.prototype.sizeRaster = true;
eq.c.DlgControl.prototype.minHeight = true;
eq.c.DlgControl.prototype.focusable = true;
eq.c.DlgControl.prototype.visible = true;
eq.c.DlgControl.prototype.sensitive = true;

/*------------------------------------------------------------------------
   static DlgControl.addClass()
   Subclassing convenience helper: Add DLG control class.
------------------------------------------------------------------------*/

eq.c.DlgControl.addClass = function(name, cls) {
   this.subClass(name, cls);
   eq.d.push(cls);
   return cls;
};

/*------------------------------------------------------------------------
   DlgControl.del()
   Delete control.
------------------------------------------------------------------------*/

eq.c.DlgControl.prototype.del = function(level) {
   this.id = undefined;
   if (this.parent) {
      if (!level)
         this.parent.childDeleted(this);
      this.parent = undefined;
      return false;
   }
   return true; // Root control deleted.
};

/*------------------------------------------------------------------------
   DlgControl.setup()
   Setup control from DLG request.
------------------------------------------------------------------------*/

eq.c.DlgControl.prototype.setup = function(dlg, rq, et) {
   var tag;
   while (tag = rq.decodeUint())
      this.setAttr(dlg, rq, tag, et);
};

/*------------------------------------------------------------------------
   static DlgControl.isSetup()
   Check if specified template level(s) exist.
   In 'modify' mode set them up if necessary.
------------------------------------------------------------------------*/

eq.c.DlgControl.isSetup = function(rq, et, levels) {
   var modify, setup = false;
   for (var i = 2; i <= levels; i++, et = et.et)
      if (et.et === undefined) {
         if (modify === undefined)
            modify = rq.mode === eq.c.DlgRq.mode.modify;
         if (!modify)
            throw new Error("setup: missing level " + i);
         et.et = {};
         setup = true;
      }
   return setup;
};

/*------------------------------------------------------------------------
   DlgControl.modify()
   Modify control from DLG request.
------------------------------------------------------------------------*/

eq.c.DlgControl.prototype.modify = function(dlg, rq, et) {
   var tag;
   while (tag = rq.decodeUint())
      this.setAttr(dlg, rq, tag, et);

   if (this.focusable && this.visible && this.sensitive) {
      var cl = et.cla = et.cla || [];
      cl.push('eq-focus');
   }
   else {
      var cl = et.clr = et.clr || [];
      cl.push('eq-focus');
      cl.push('eq-noenter');
      if (et.cla !== undefined) {
         var cl = et.cla, i = cl.indexOf('eq-noenter');
         if (i !== -1)
            cl.splice(i, 1);
      }
   }
};

/*------------------------------------------------------------------------
   DlgControl.setAttr()
   Set control attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.DlgControl.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {

      case rt.cssid:
         this.id = et.id = rq.decodeString();
         break;

      case rt.cssclsDel: {
         var sv, cl = et.clr = et.clr || [];
         while ((sv = rq.decodeString()).length)
            cl.push(sv);
         break;
      }

      case rt.cssclsAdd: {
         var sv, cl = et.cla = et.cla || [];
         while ((sv = rq.decodeString()).length)
            cl.push(sv);
         break;
      }

      case rt.raster: {
         var iv = rq.decodeUint();
         this.posRaster = !!(iv & 1);
         this.sizeRaster = !!(iv & 2);
         break;
      }

      case rt.posSize: {
         var
            api = eq.api,
            st = et.st = et.st || {},
            iv;
         switch (rq.decodeInt()) { // xauto
            case -1:
               iv = rq.decodeInt(); // xright
               if (this.posRaster)
                  st['right'] = api.cssWidth(iv);
               else
                  st['right'] = iv ? iv + 'px' : 0;
               if (iv = rq.decodeInt()) { // width
                  if (this.sizeRaster)
                     st['width'] = api.cssWidth(iv);
                  else
                     st['width'] = iv + 'px';
               }
               else if (this.minWidth)
                  st['width'] = api.cssWidth(1);
               else
                  st['width'] = null;
               st['left'] = null;
               break;
            case 0:
               iv = rq.decodeInt(); // xleft
               if (this.posRaster)
                  st['left'] = api.cssWidth(iv);
               else
                  st['left'] = iv ? iv + 'px' : 0;
               iv = rq.decodeInt(); // xright
               if (this.posRaster)
                  st['right'] = api.cssWidth(iv);
               else
                  st['right'] = iv ? iv + 'px' : 0;
               st['width'] = null;
               break;
            case 1:
               iv = rq.decodeInt(); // xleft
               if (this.posRaster)
                  st['left'] = api.cssWidth(iv);
               else
                  st['left'] = iv ? iv + 'px' : 0;
               if (iv = rq.decodeInt()) { // width
                  if (this.sizeRaster)
                     st['width'] = api.cssWidth(iv);
                  else
                     st['width'] = iv + 'px';
               }
               else if (this.minWidth)
                  st['width'] = api.cssWidth(1);
               else
                  st['width'] = null;
               st['right'] = null;
               break;
         }

         switch (rq.decodeInt()) { // yauto
            case -1:
               iv = rq.decodeInt(); // ybottom
               if (this.posRaster)
                  st['bottom'] = api.cssHeight(iv);
               else
                  st['bottom'] = iv ? iv + 'px' : 0;
               if (iv = rq.decodeInt()) { // height
                  if (this.sizeRaster)
                     st['height'] = api.cssHeight(iv);
                  else
                     st['height'] = iv + 'px';
               }
               else if (this.minHeight)
                  st['height'] = api.cssHeight(1);
               else
                  st['height'] = null;
               st['top'] = null;
               break;
            case 0:
               iv = rq.decodeInt(); // ytop
               if (this.posRaster)
                  st['top'] = api.cssHeight(iv);
               else
                  st['top'] = iv ? iv + 'px' : 0;
               iv = rq.decodeInt(); // ybottom
               if (this.posRaster)
                  st['bottom'] = api.cssHeight(iv);
               else
                  st['bottom'] = iv ? iv + 'px' : 0;
               st['height'] = null;
               break;
            case 1:
               iv = rq.decodeInt(); // ytop
               if (this.posRaster)
                  st['top'] = api.cssHeight(iv);
               else
                  st['top'] = iv ? iv + 'px' : 0;
               if (iv = rq.decodeInt()) { // height
                  if (this.sizeRaster)
                     st['height'] = api.cssHeight(iv);
                  else
                     st['height'] = iv + 'px';
               }
               else if (this.minHeight)
                  st['height'] = api.cssHeight(1);
               else
                  st['height'] = null;
               st['bottom'] = null;
               break;
         }

         break;
      }

      case rt.size: {
         var
            api = eq.api,
            st = et.st = et.st || {},
            iv;
         if (iv = rq.decodeInt()) { // width
            if (this.sizeRaster)
               st['width'] = api.cssWidth(iv);
            else
               st['width'] = iv ? iv + 'px' : 0;
         }
         else if (this.minWidth)
            st['width'] = api.cssWidth(1);
         else
            st['width'] = null;
         if (iv = rq.decodeInt()) { // height
            if (this.sizeRaster)
               st['height'] = api.cssHeight(iv);
            else
               st['height'] = iv ? iv + 'px' : 0;
         }
         else if (this.minHeight)
            st['height'] = api.cssHeight(1);
         else
            st['height'] = null;
         break;
      }

      case rt.border: {
         var cl;
         if (rq.decodeUint())
            cl = et.cla = et.cla || [];
         else
            cl = et.clr = et.clr || [];
         cl.push('eq-border');
         break;
      }

      case rt.rule: {
         var
            rule = rq.decodeInt(),
            cl;
         if (rule)
            this.rule = rule;
         else
            delete this.rule;
         this.setRuleAttr(dlg, rq, et, rule);
         if (rule === -1) {
            cl = et.cla = et.cla || [];
            this.focusable = false;
            // Invalidate tab order.
            dlg.tabOrderValid = false;
         }
         else if (this.hasOwnProperty('focusable')) {
            cl = et.clr = et.clr || [];
            if (this.parent && !this.parent.containerFocusable())
               this.focusable = false;
            else
               delete this.focusable;
            // Invalidate tab order.
            dlg.tabOrderValid = false;
         }
         if (cl !== undefined)
            cl.push('eq-help');
         break;
      }

      case rt.tabonenter: {
         var cl;
         if (rq.decodeUint())
            cl = et.clr = et.clr || [];
         else
            cl = et.cla = et.cla || [];
         cl.push('eq-noenter');
         break;
      }

      case rt.visible: {
         var cl, st = et.st = et.st || {};
         if (rq.decodeUint()) {
            cl = et.clr = et.clr || [];
            st['display'] = null;
            delete this.visible;
         }
         else {
            cl = et.cla = et.cla || [];
            st['display'] = 'none';
            this.visible = false;
         }
         cl.push('eq-hidden');
         // Invalidate tab order.
         dlg.tabOrderValid = false;
         break;
      }

      case rt.sensitive: {
         var cl;
         if (rq.decodeUint()) {
            cl = et.clr = et.clr || [];
            delete this.sensitive;
         }
         else {
            cl = et.cla = et.cla || [];
            this.sensitive = false;
         }
         cl.push('eq-disabled');
         this.setSensitive(dlg, rq, et);
         // Invalidate tab order.
         dlg.tabOrderValid = false;
         break;
      }

      case rt.inactive: {
         if (rq.decodeUint())
            this.inactive = true;
         else
            delete this.inactive;
         // Invalidate tab order.
         dlg.tabOrderValid = false;
         break;
      }

      case rt.tabStop: {
         var cl;
         if (rq.decodeUint())
            cl = et.clr = et.clr || [];
         else
            cl = et.cla = et.cla || [];
         cl.push('eq-notab');
         break;
      }

      case rt.bgColor: {
         var
            sv = rq.decodeString(),
            st = et.st = et.st || {},
            cl;
         if (sv.length) {
            st['background-color'] = sv;
            cl = et.cla = et.cla || [];
         }
         else {
            st['background-color'] = null;
            cl = et.clr = et.clr || [];
         }
         cl.push('eq-bgc');
         break;
      }

      case rt.fgColor: {
         var
            sv = rq.decodeString(),
            st = et.st = et.st || {},
            cl;
         if (sv.length) {
            st['color'] = sv;
            cl = et.cla = et.cla || [];
         }
         else {
            st['color'] = null;
            cl = et.clr = et.clr || [];
         }
         cl.push('eq-fgc');
         break;
      }

      default:
         throw new Error("DlgControl.setAttr(" +
                         this.id + ") failed: tag " + tag);
   }
};

/*------------------------------------------------------------------------
   DlgControl.setRuleAttr()
   Set 'rule' attribute from DLG request.
------------------------------------------------------------------------*/

eq.c.DlgControl.prototype.setRuleAttr = function(dlg, rq, et) {};

/*------------------------------------------------------------------------
   DlgControl.setSensitive()
   Set element-specific 'sensitive' properties.
------------------------------------------------------------------------*/

eq.c.DlgControl.prototype.setSensitive = function(dlg, rq, et) {};

/*------------------------------------------------------------------------
   DlgControl.findById()
   Find control by id.
------------------------------------------------------------------------*/

eq.c.DlgControl.prototype.findById = function(id) {
   return id === this.id ? this : null;
};

/*------------------------------------------------------------------------
   DlgControl.updateTabOrder()
   Fill current tab order array.
------------------------------------------------------------------------*/

eq.c.DlgControl.prototype.updateTabOrder = function(tabOrder) {
   if (this.visible && this.sensitive && !this.inactive && this.focusable)
      tabOrder.push(this.id);
}

/*------------------------------------------------------------------------
   DlgControl.updateTabOrder()
   Act upon main thread requesting to update tab order.
------------------------------------------------------------------------*/

eq.c.DlgControl.prototype.onUpdateTabOrder = function(arg) {};

/*========================================================================
   class DlgContainer extends DlgControl
   DLG container base class.
========================================================================*/

eq.c.DlgControl.subClass('DlgContainer', function() {
   eq.c.DlgControl.call(this);

   this.ch = []; // Children
});

eq.c.DlgContainer.prototype.focusable = false;

/*------------------------------------------------------------------------
   static DlgContainer.addClass()
   Subclassing convenience helper: Add DLG container class.
------------------------------------------------------------------------*/

eq.c.DlgContainer.addClass = function(name, cls) {
   return eq.c.DlgControl.addClass.call(this, name, cls);
};

/*------------------------------------------------------------------------
   DlgContainer.addChild()
   Add child control.
------------------------------------------------------------------------*/

eq.c.DlgContainer.prototype.addChild = function(child) {
   this.ch.push(child);
   child.parent = this;
};

/*------------------------------------------------------------------------
   DlgContainer.childDeleted()
   Child control deleted.
------------------------------------------------------------------------*/

eq.c.DlgContainer.prototype.childDeleted = function(child) {
   var i = this.ch.indexOf(child);
   if (i === -1)
      throw new Error("DlgContainer.childDeleted(" +
                      this.id + ", " + child.id + ") failed: not found");

   this.ch.splice(i, 1);
}

/*------------------------------------------------------------------------
   DlgContainer.del()
   Delete container.
------------------------------------------------------------------------*/

eq.c.DlgContainer.prototype.del = function(level) {
   if (level === undefined)
      level = 0;

   for (var i = this.ch.length; --i >= 0;)
      this.ch[i].del(level + 1);
   this.ch.length = 0;

   return eq.c.DlgControl.prototype.del.call(this, level);
};

/*------------------------------------------------------------------------
   DlgContainer.setup()
   Setup container from DLG request, create children if any.
------------------------------------------------------------------------*/

eq.c.DlgContainer.prototype.setup = function(dlg, rq, et) {
   var tag;
   while (tag = rq.decodeUint()) {
      if (tag === eq.RqTag.child) {
         var tmp = et;
         while (tmp.et !== undefined)
            tmp = tmp.et;
         dlg.addChild(rq, this, tmp);
      }
      else
         this.setAttr(dlg, rq, tag, et);
   }
};

/*------------------------------------------------------------------------
   DlgContainer.findById()
   Find control by id.
------------------------------------------------------------------------*/

eq.c.DlgContainer.prototype.findById = function(id) {
   if (id === this.id)
      return this;

   for (var i = 0, l = this.ch.length; i < l; i++) {
      var found = this.ch[i].findById(id);
      if (found)
         return found;
   }

   return null;
};

/*------------------------------------------------------------------------
   DlgContainer.containerFocusable()
   Check 'focusable' based on container.
------------------------------------------------------------------------*/

eq.c.DlgContainer.prototype.containerFocusable = function() {
   if (this.parent)
      return this.parent.containerFocusable();
   return true;
};

/*------------------------------------------------------------------------
   DlgContainer.updateTabOrder()
   Fill current tab order array.
------------------------------------------------------------------------*/

eq.c.DlgContainer.prototype.updateTabOrder = function(tabOrder) {
   if (this.visible && this.sensitive && !this.inactive) {
      if (this.focusable)
         tabOrder.push(this.id);
      for (var i = 0, l = this.ch.length; i < l; i++)
         this.ch[i].updateTabOrder(tabOrder);
   }
}

/** @preserve $Id: 10-checkbox.js,v 28.15 2021/07/13 13:18:15 mike Exp $
*//*======================================================================
   class CheckBox extends DlgControl
   DLG CheckBox class.
========================================================================*/

eq.c.DlgControl.addClass('CheckBox', function() {
   eq.c.DlgControl.call(this);
});

/*------------------------------------------------------------------------
   CheckBox.setup()
   Setup CheckBox from DLG request.
------------------------------------------------------------------------*/

eq.c.CheckBox.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-checkbox', 'eq-control' ];
   et.et = {
      ch : [
         { id : '*' },
         { id : '*' }
      ]
   };
   eq.c.DlgControl.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   CheckBox.setAttr()
   Set CheckBox attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.CheckBox.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.textPos:
      case rt.active:
      case rt.text:
         if (eq.c.DlgControl.isSetup(rq, et, 2))
            et.et.ch = et.et.ch || [{},{}];
         break;

      default:
         eq.c.DlgControl.prototype.setAttr.call(this, dlg, rq, tag, et);
         return;
   }

   switch (tag) {
      case rt.textPos: {
         // Button index, 0 = label right, 1 = label left.
         this.iBtn = rq.decodeUint();
         // Button.
         var ch = et.et.ch[this.iBtn];
         ch.tg = 'INPUT';
         ch.at = { 'type' : 'checkbox' };
         ch.ev = {
            c : [ 'click' ]
         };
         // Label.
         ch = et.et.ch[this.iBtn ? 0 : 1];
         ch.tg = 'LABEL';
         ch.at = { 'for' : et.id + '-' + (this.iBtn + 1) };
         break;
      }

      case rt.active: {
         var ch = et.et.ch[this.iBtn];
         ch.at = ch.at || {};
         ch.at.checked = rq.decodeUint() !== 0;
         break;
      }

      case rt.text: {
         var ch = et.et.ch[this.iBtn ? 0 : 1];
         ch.hca = rq.decodeString();
         break;
      }
   }
};

/*------------------------------------------------------------------------
   CheckBox.setRuleAttr()
   Set 'rule' attribute from DLG request.
------------------------------------------------------------------------*/

eq.c.CheckBox.prototype.setRuleAttr = function(dlg, rq, et, rule) {
   if (eq.c.DlgControl.isSetup(rq, et, 2))
      et.et.ch = et.et.ch || [{},{}];
   var ch = et.et.ch[this.iBtn], ev = ch.ev = ch.ev || {};
   // Element is focusable and can submit,
   // use 'mousedown', not 'click'.
   ev.a = rule ? [ 'mousedown' ] : null;
};

/*------------------------------------------------------------------------
   CheckBox.setSensitive()
   Set element-specific 'sensitive' properties.
------------------------------------------------------------------------*/

eq.c.CheckBox.prototype.setSensitive = function(dlg, rq, et) {
   if (eq.c.DlgControl.isSetup(rq, et, 2))
      et.et.ch = et.et.ch || [{},{}];
   var ch = et.et.ch[this.iBtn], at = ch.at = ch.at || {};
   at.disabled = this.sensitive ? null : true;
};

/** @preserve $Id: 10-combobox.js,v 28.6 2021/06/02 15:43:51 mike Exp $
*//*======================================================================
   class ComboBox extends DlgControl
   DLG ComboBox class.
========================================================================*/

eq.c.DlgControl.addClass('ComboBox', function() {
   eq.c.DlgControl.call(this);

   // Default height.
   this.defaultHeight = 8;
   var api = eq.api;
   this.defaultPxHeight = api.pxHeight(this.defaultHeight);
   this.defaultCssHeight = api.cssHeight(this.defaultHeight);
});

/*------------------------------------------------------------------------
   ComboBox.setup()
   Setup ComboBox from DLG request.
------------------------------------------------------------------------*/

eq.c.ComboBox.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-combobox', 'eq-control', 'eq-border' ];
   et.et = {
      tg  : 'INPUT',
      at  : {
         'type' : 'text',
         'autocorrect' : 'off',
         'spellcheck' : 'false'
      },
      ev  : {
         c : [ 'input', 'mousedown' ]
      }
   };
   eq.c.DlgControl.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   ComboBox.setAttr()
   Set ComboBox attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.ComboBox.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.items:
      case rt.boxSize:
         break;
      case rt.readonly:
      case rt.content:
      case rt.align:
         eq.c.DlgControl.isSetup(rq, et, 2);
         break;

      default:
         eq.c.DlgControl.prototype.setAttr.call(this, dlg, rq, tag, et);
         return;
   }

   switch (tag) {
      case rt.items: {
         var
            l = rq.decodeUint(),
            rt = et.rt = et.rt || [],
            v = [];
         v.length = l;
         for (var i = 0; i < l; i++)
            v[i] = rq.decodeString();
         rt.push({ t : tag, v : v });
         break;
      }

      case rt.boxSize: {
         var
            w = rq.decodeUint(),
            h = rq.decodeUint(),
            rt = et.rt = et.rt || [],
            cw, ch, mh;
         if (this.sizeRaster) {
            var api = eq.api;
            if (w)
               cw = api.cssWidth(w);
            if (h && h > this.defaultHeight)
               ch = api.cssHeight(h);
            else
               ch = this.defaultCssHeight;
         }
         else {
            if (w)
               cw = w + 'px';
            if (h && h > this.defaultPxHeight)
               ch = h + 'px';
            else
               ch = this.defaultCssHeight;
         }
         rt.push({ t : tag, w : cw, h : ch });
         break;
      }

      case rt.readonly: {
         var
            cl,
            at = et.et.at = et.et.at || {};
         if (rq.decodeUint()) {
            cl = et.cla = et.cla || [];
            at.readonly = true;
            this.focusable = false;
         }
         else {
            cl = et.clr = et.clr || [];
            at.readonly = null;
            delete this.focusable;
         }
         cl.push('eq-readonly');
         // Invalidate tab order.
         dlg.tabOrderValid = false;
         break;
      }

      case rt.content:
         et.et.ec = rq.decodeString();
         break;

      case rt.align: {
         var st = et.et.st = et.et.st || {};
         switch (rq.decodeUint()) {
            case 0:
               st['text-align'] = null;
               break;
            case 1:
               st['text-align'] = 'center';
               break;
            case 2:
               st['text-align'] = 'right';
         }
         break;
      }
   }
};

/*------------------------------------------------------------------------
   ComboBox.setRuleAttr()
   Set 'rule' attribute from DLG request.
------------------------------------------------------------------------*/

eq.c.ComboBox.prototype.setRuleAttr = function(dlg, rq, et, rule) {
   eq.c.DlgControl.isSetup(rq, et, 2);
   var ev = et.et.ev = et.et.ev || {};
   ev.a = rule ? [ 'blur' ] : null;
};

/*------------------------------------------------------------------------
   ComboBox.setSensitive()
   Set element-specific 'sensitive' properties.
------------------------------------------------------------------------*/

eq.c.ComboBox.prototype.setSensitive = function(dlg, rq, et) {
   eq.c.DlgControl.isSetup(rq, et, 2);
   var at = et.et.at = et.et.at || {};
   at.disabled = this.sensitive ? null : true;
};

/** @preserve $Id: 10-dialog.js,v 28.9 2021/07/12 09:25:23 mike Exp $
*//*======================================================================
   class Dialog extends DlgContainer
   DLG Dialog class.
========================================================================*/

eq.c.DlgContainer.addClass('Dialog', function() {
   eq.c.DlgContainer.call(this);
});

/*------------------------------------------------------------------------
   Dialog.setup()
   Setup Dialog from DLG request.
------------------------------------------------------------------------*/

eq.c.Dialog.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-dialog', 'eq-border' ];
   et.et = {
      cla : [ 'eq-pane' ],
      et : {
         cla : [ 'eq-container' ]
      }
   };
   eq.c.DlgContainer.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   Dialog.setAttr()
   Set Dialog attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.Dialog.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.dialogPosSize: {
         var
            d = et.dlg = et.dlg || {},
            l = d.l = d.l || {},
            cla = et.cla = et.cla || [],
            clr = et.clr = et.clr || [],
            st = et.st = et.st || {},
            iv, pst;

         d.resizable = rq.decodeUint();

         if ((iv = rq.decodeInt()) === -1) // xleft
            l.x = -1;
         else if (this.posRaster)
            l.x = eq.api.cssWidth(iv);
         else
            l.x = iv ? iv + 'px' : 0;
         if ((iv = rq.decodeInt()) === -1) // ytop
            l.y = -1;
         else if (this.posRaster)
            l.y = eq.api.cssHeight(iv);
         else
            l.y = iv ? iv + 'px' : 0;
         iv = rq.decodeInt(); // width
         if (this.sizeRaster)
            l.w = eq.api.cssWidth(iv);
         else
            l.w = iv ? iv + 'px' : 0;
         iv = rq.decodeInt(); // height
         if (this.sizeRaster)
            l.h = eq.api.cssHeight(iv);
         else
            l.h = iv ? iv + 'px' : 0;

         if (l.x === -1 && l.y === -1) {
            cla.push('eq-center');
            clr.push('eq-hcenter');
            clr.push('eq-vcenter');
            st['top'] = null;
            st['left'] = null;
         }
         else if (l.x === -1) {
            cla.push('eq-hcenter');
            clr.push('eq-center');
            clr.push('eq-vcenter');
            st['top'] = l.y;
            st['left'] = null;
         }
         else if (l.y === -1) {
            cla.push('eq-vcenter');
            clr.push('eq-center');
            clr.push('eq-hcenter');
            st['top'] = null;
            st['left'] = l.x;
         }
         else {
            st['top'] = l.y;
            st['left'] = l.x;
         }
         st['right'] = null;
         st['bottom'] = null;

         eq.c.DlgControl.isSetup(rq, et, 2);
         pst = et.et.st = et.et.st || {};
         pst['width'] = l.w;
         pst['height'] = l.h;
         break;
      }

      case rt.posSize: {
         var
            d = et.dlg = et.dlg || {},
            cl = et.clr = et.clr || [],
            st = et.st = et.st || {},
            iv, pst;
         d.l = null;
         cl.push('eq-center');
         cl.push('eq-hcenter');
         cl.push('eq-vcenter');

         eq.c.DlgControl.isSetup(rq, et, 2);
         pst = et.et.st = et.et.st || {};
         switch (rq.decodeInt()) { // xauto
            case -1:
               iv = rq.decodeInt(); // xright
               if (this.posRaster)
                  st['right'] = eq.api.cssWidth(iv);
               else
                  st['right'] = iv ? iv + 'px' : 0;
               if (iv = rq.decodeInt()) { // width
                  if (this.sizeRaster)
                     pst['width'] = eq.api.cssWidth(iv);
                  else
                     pst['width'] = iv + 'px';
               }
               else if (this.minWidth)
                  pst['width'] = eq.api.cssWidth(1);
               st['left'] = null;
               break;
            case 0:
               iv = rq.decodeInt(); // xleft
               if (this.posRaster)
                  st['left'] = eq.api.cssWidth(iv);
               else
                  st['left'] = iv ? iv + 'px' : 0;
               iv = rq.decodeInt(); // xright
               if (this.posRaster)
                  st['right'] = eq.api.cssWidth(iv);
               else
                  st['right'] = iv ? iv + 'px' : 0;
               pst['width'] = null;
               break;
         }

         switch (rq.decodeInt()) { // yauto
            case -1:
               iv = rq.decodeInt(); // ybottom
               if (this.posRaster)
                  st['bottom'] = eq.api.cssHeight(iv);
               else
                  st['bottom'] = iv ? iv + 'px' : 0;
               if (iv = rq.decodeInt()) { // height
                  if (this.sizeRaster)
                     pst['height'] = eq.api.cssHeight(iv);
                  else
                     pst['height'] = iv + 'px';
               }
               else
                  pst['height'] = eq.api.cssHeight(1);
               st['top'] = null;
               break;
            case 0:
               iv = rq.decodeInt(); // ytop
               if (this.posRaster)
                  st['top'] = eq.api.cssHeight(iv);
               else
                  st['top'] = iv ? iv + 'px' : 0;
               iv = rq.decodeInt(); // ybottom
               if (this.posRaster)
                  st['bottom'] = eq.api.cssHeight(iv);
               else
                  st['bottom'] = iv ? iv + 'px' : 0;
               pst['height'] = null;
               break;
         }

         break;
      }

      case rt.fKey: {
         var
            d = et.dlg = et.dlg || {},
            fkey = [],
            l = rq.decodeUint();
         fkey.length = l;
         for (var i = 0; i < l; i++)
            fkey[i] = rq.decodeInt();
         d.f = fkey;
         break;
      }

      case rt.text: {
         var d = et.dlg = et.dlg || {};
         d.t = rq.decodeString();
         break;
      }

      default:
         eq.c.DlgContainer.prototype.setAttr.call(this, dlg, rq, tag, et);
         return;
   }
};

/** @preserve $Id: 10-edittext.js,v 28.25 2021/06/14 12:43:23 mike Exp $
*//*======================================================================
   class EditText extends DlgControl
   DLG EditText class.
========================================================================*/

eq.c.DlgControl.addClass('EditText', function() {
   eq.c.DlgControl.call(this);
});

/*------------------------------------------------------------------------
   EditText.setup()
   Setup EditText from DLG request.
------------------------------------------------------------------------*/

eq.c.EditText.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-edittext', 'eq-control' ];
   et.et = {
      ev  : {
         c : [ 'input', 'blur' ]
      }
   };
   eq.c.DlgControl.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   EditText.setAttr()
   Set EditText attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.EditText.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.multiline:
      case rt.readonly:
      case rt.content:
      case rt.ispassword:
      case rt.align:
         eq.c.DlgControl.isSetup(rq, et, 2);
         break;

      default:
         eq.c.DlgControl.prototype.setAttr.call(this, dlg, rq, tag, et);
         return;
   }

   switch (tag) {
      case rt.multiline: {
         var
            tt = et.et,
            at = tt.at = tt.at || {},
            cl;
         if (this.multiline = rq.decodeUint()) {
            tt.tg = 'TEXTAREA';
            cl = et.cla = et.cla || [];
            // Disable auto-completion, auto-correction, spell-checking.
            at.autocapitalize = 'none';
            at.autocomplete = 'off';
            // TODO: wrap
            at.wrap = 'off'; // IE
         }
         else {
            tt.tg = 'INPUT';
            cl = et.clr = et.clr || [];
            at.type = this.ispassword ? 'password' : 'text';
         }
         // Disable auto-correction.
         at.autocorrect = 'off';
         at.spellcheck = 'false';
         cl.push('eq-noenter');
         break;
      }

      case rt.readonly: {
         var
            cl,
            at = et.et.at = et.et.at || {};
         if (rq.decodeUint()) {
            cl = et.cla = et.cla || [];
            at.readonly = true;
            this.focusable = false;
         }
         else {
            cl = et.clr = et.clr || [];
            at.readonly = null;
            delete this.focusable;
         }
         cl.push('eq-readonly');
         // Invalidate tab order.
         dlg.tabOrderValid = false;
         break;
      }

      case rt.content:
         et.et.ec = rq.decodeString();
         break;

      case rt.ispassword:
         this.ispassword = rq.decodeUint();
         if (!this.multiline) {
            var at = et.et.at = et.et.at || {};
            at.type = this.ispassword ? 'password' : 'text';
         }
         break;

      case rt.align: {
         var st = et.et.st = et.et.st || {};
         switch (rq.decodeUint()) {
            case 0:
               st['text-align'] = null;
               break;
            case 1:
               st['text-align'] = 'center';
               break;
            case 2:
               st['text-align'] = 'right';
         }
         break;
      }
   }
};

/*------------------------------------------------------------------------
   EditText.setRuleAttr()
   Set 'rule' attribute from DLG request.
------------------------------------------------------------------------*/

eq.c.EditText.prototype.setRuleAttr = function(dlg, rq, et, rule) {
   eq.c.DlgControl.isSetup(rq, et, 2);
   var ev = et.et.ev = et.et.ev || {};
   ev.a = rule ? [ 'blur' ] : null;
};

/*------------------------------------------------------------------------
   EditText.setSensitive()
   Set element-specific 'sensitive' properties.
------------------------------------------------------------------------*/

eq.c.EditText.prototype.setSensitive = function(dlg, rq, et) {
   eq.c.DlgControl.isSetup(rq, et, 2);
   var at = et.et.at = et.et.at || {};
   at.disabled = this.sensitive ? null : true;
};

/** @preserve $Id: 10-groupbox.js,v 28.7 2021/06/01 18:12:29 mike Exp $
*//*======================================================================
   class GroupBox extends DlgContainer
   DLG GroupBox class.
========================================================================*/

eq.c.DlgContainer.addClass('GroupBox', function() {
   eq.c.DlgContainer.call(this);
});

/*------------------------------------------------------------------------
   GroupBox.setup()
   Setup GroupBox from DLG request.
------------------------------------------------------------------------*/

eq.c.GroupBox.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-groupbox', 'eq-border' ];
   et.et = {
      cla : [ 'eq-pane' ],
      et  : {
         cla : [ 'eq-container' ]
      }
   };
   eq.c.DlgContainer.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   GroupBox.setAttr()
   Set GroupBox attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.GroupBox.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.vWidth:
      case rt.vHeight:
         eq.c.DlgControl.isSetup(rq, et, 2);
         break;

      default:
         eq.c.DlgContainer.prototype.setAttr.call(this, dlg, rq, tag, et);
         return;
   }

   switch (tag) {
      case rt.vWidth: {
         var
            iv = rq.decodeUint(),
            st = et.et.st = et.et.st || {};
            if (iv) {
               if (this.sizeRaster)
                  st['width'] = eq.api.cssWidth(iv);
               else
                  st['width'] = iv + 'px';
            }
            else
               st['width'] = '';
         break;
      }

      case rt.vHeight: {
         var
            iv = rq.decodeUint(),
            st = et.et.st = et.et.st || {};
            if (iv) {
               if (this.sizeRaster)
                  st['height'] = eq.api.cssHeight(iv);
               else
                  st['height'] = iv + 'px';
            }
            else
               st['height'] = '';
         break;
      }
   }
};

/** @preserve $Id: 10-htmlview.js,v 28.4 2021/07/13 13:16:31 mike Exp $
*//*======================================================================
   class HtmlView extends DlgControl
   DLG HtmlView class.
========================================================================*/

eq.c.DlgControl.addClass('HtmlView', function() {
   eq.c.DlgControl.call(this);
});

eq.c.HtmlView.prototype.focusable = false;

/*------------------------------------------------------------------------
   HtmlView.setup()
   Setup HtmlView from DLG request.
------------------------------------------------------------------------*/

eq.c.HtmlView.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-htmlview', 'eq-control', 'eq-border' ];
   et.ev = { c : [ 'change' ] };
   eq.c.DlgControl.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   HtmlView.setAttr()
   Set HtmlView attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.HtmlView.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.url: {
         var rt = et.rt = et.rt || [];
         rt.push({
            t : tag,
            v : eq.api.makeUrl(rq.decodeString())
         });
         break;
      }

      case rt.content: {
         var rt = et.rt = et.rt || [];
         rt.push({
            t : tag,
            ty : rq.decodeUint(),
            v : rq.decodeString()
         });
         break;
      }

      default:
         eq.c.DlgControl.prototype.setAttr.call(this, dlg, rq, tag, et);
         return;
   }
};

/*------------------------------------------------------------------------
   HtmlView.setRuleAttr()
   Set 'rule' attribute from DLG request.
------------------------------------------------------------------------*/

eq.c.HtmlView.prototype.setRuleAttr = function(dlg, rq, et, rule) {
   var cl;
   if (rule)
      cl = et.cla = et.cla || [];
   else
      cl = et.clr = et.clr || [];
   cl.push('eq-rule');
};

/** @preserve $Id: 10-image.js,v 28.8 2021/07/13 13:16:31 mike Exp $
*//*======================================================================
   class Image extends DlgControl
   DLG Image class.
========================================================================*/

eq.c.DlgControl.addClass('Image', function() {
   eq.c.DlgControl.call(this);
});

eq.c.Image.prototype.focusable = false;

/*------------------------------------------------------------------------
   Image.setup()
   Setup Image from DLG request.
------------------------------------------------------------------------*/

eq.c.Image.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-image', 'eq-control', 'eq-border' ];
   et.et = {
      tg : 'IMG'
   };
   eq.c.DlgControl.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   Image.setAttr()
   Set Image attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.Image.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.url:
         eq.c.DlgControl.isSetup(rq, et, 2);
         break;

      default:
         eq.c.DlgControl.prototype.setAttr.call(this, dlg, rq, tag, et);
         return;
   }

   switch (tag) {
      case rt.url: {
         var sv = eq.api.makeUrl(rq.decodeString());
         et.et.at = et.et.at || {};
         et.et.at['src'] = sv.length ? sv : null;
         break;
      }
   }
};

/** @preserve $Id: 10-listbox.js,v 28.17 2021/07/04 16:13:09 mike Exp $
*//*======================================================================
   class ListBox extends DlgContainer
   DLG ListBox class.
========================================================================*/

eq.c.DlgContainer.addClass('ListBox', function() {
   eq.c.DlgContainer.call(this);
});

eq.c.ListBox.prototype.focusable = true;

/*------------------------------------------------------------------------
   ListBox.setup()
   Setup ListBox from DLG request.
------------------------------------------------------------------------*/

eq.c.ListBox.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-listbox', 'eq-control', 'eq-border' ];
   et.ev = {
      // Element is focusable and can submit,
      // use 'mousedown', not 'click'.
      c : [ 'mousedown', 'scroll' ]
   };
   eq.c.DlgControl.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   ListBox.setAttr()
   Set ListBox attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.ListBox.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.columns: {
         var
            l = rq.decodeUint(),
            rt = et.rt = et.rt || [];
         if (l === 0) {
            // Single-column ListBox.
            this.ci = undefined;
            this.cs = undefined;
            rt.push({ t : tag });
         }
         else {
            // Multi-column ListBox.
            var
               ci = [],
               cw = [],
               ct = [];
            this.ci = ci;
            this.cs = rq.decodeString();
            ci.length = cw.length = ct.length = l;
            for (var i = 0; i < l; i++) {
               ci[i] = rq.decodeUint();
               var
                  f = rq.decodeUint(),
                  w = rq.decodeUint();
               if (f & 1) // sizeRaster
                  cw[i] = eq.api.cssWidth(w);
               else
                  cw[i] = w + 'px';
               ct[i] = rq.decodeString();
            }
            rt.push({
               t  : tag,
               cw : cw,
               ct : ct
            });
         }
         break;
      }

      case rt.content: {
         var
            l = rq.decodeUint(),
            rt = et.rt = et.rt || [],
            v = [];
         v.length = l;
         for (var i = 0; i < l; i++)
            v[i] = rq.decodeString();
         if (this.ci === undefined)
            rt.push({ t : tag, v : v });
         else
            rt.push({
               t  : tag,
               v  : v,
               ci : this.ci,
               cs : this.cs
            });
         break;
      }

      case rt.activeLine:
      case rt.topItem: {
         var rt = et.rt = et.rt || [];
         rt.push({ t : tag, v : rq.decodeUint() });
         break;
      }

      case rt.align: {
         var
            l = rq.decodeUint(),
            rt = et.rt = et.rt || [],
            ca = [];
         ca.length = l;
         for (var i = 0; i < l; i++)
            ca[i] = rq.decodeUint();
         rt.push({
            t  : tag,
            ca : ca
         });
         break;
      }

      case rt.ruleMode:
         this.ruleMode = rq.decodeUint();
         break;

      default:
         eq.c.DlgControl.prototype.setAttr.call(this, dlg, rq, tag, et);
   }
};

/*------------------------------------------------------------------------
   ListBox.setRuleAttr()
   Set 'rule' attribute from DLG request.
------------------------------------------------------------------------*/

eq.c.ListBox.prototype.setRuleAttr = function(dlg, rq, et, rule) {
   var ev = et.ev = et.ev || {};
   if (rule) {
      // Element is focusable and can submit,
      // use 'mousedown', not 'click'.
      ev.a = [ 'mousedown', 'keydown' ];
      switch (this.ruleMode) {
         case 1:
            ev.m = 1;
            ev.f = null;
            break;
         case 2:
            ev.m = 2;
            ev.f = 1;
            break;
         case 3:
            ev.m = 1;
            ev.f = 1;
            break;
         default:
            ev.m = 2;
            ev.f = null;
      }
   }
   else {
      ev.a = null;
      ev.m = null;
      ev.f = null;
   }
};

/** @preserve $Id: 10-menubar.js,v 28.4 2021/05/07 15:05:36 mike Exp $
*//*======================================================================
   class MenuBar extends DlgContainer
   DLG MenuBar class.
========================================================================*/

eq.c.DlgContainer.addClass('MenuBar', function() {
   eq.c.DlgContainer.call(this);
});

/*------------------------------------------------------------------------
   MenuBar.setup()
   Setup MenuBar from DLG request.
------------------------------------------------------------------------*/

eq.c.MenuBar.prototype.setup = function(dlg, rq, et) {
   et.tg = 'UL';
   et.cla = [ 'eq-menubar', 'eq-control' ];
   eq.c.DlgContainer.prototype.setup.call(this, dlg, rq, et);
};

/** @preserve $Id: 10-menuitem.js,v 28.6 2021/06/14 12:43:08 mike Exp $
*//*======================================================================
   class MenuItem extends DlgControl
   DLG MenuItem class.
========================================================================*/

eq.c.DlgControl.addClass('MenuItem', function() {
   eq.c.DlgControl.call(this);
});

eq.c.MenuItem.prototype.focusable = false;

/*------------------------------------------------------------------------
   MenuItem.setup()
   Setup MenuItem from DLG request.
------------------------------------------------------------------------*/

eq.c.MenuItem.prototype.setup = function(dlg, rq, et) {
   et.tg = 'LI';
   et.cla = [ 'eq-item', 'eq-control' ];
   et.ev = { c : [ 'mousedown' ] };
   eq.c.DlgControl.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   MenuItem.setAttr()
   Set MenuItem attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.MenuItem.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.text: {
         var rt = et.rt = et.rt || [];
         rt.push({ t : tag, v : rq.decodeString() });
         break;
      }

      default:
         eq.c.DlgControl.prototype.setAttr.call(this, dlg, rq, tag, et);
   }
};

/*------------------------------------------------------------------------
   MenuItem.setRuleAttr()
   Set 'rule' attribute from DLG request.
------------------------------------------------------------------------*/

eq.c.MenuItem.prototype.setRuleAttr = function(dlg, rq, et, rule) {
   var ev = et.ev = et.ev || {};
   ev.a = rule ? [ 'click' ] : null;
};

/** @preserve $Id: 10-menu.js,v 28.4 2021/04/30 16:00:35 mike Exp $
*//*======================================================================
   class Menu extends DlgContainer
   DLG Menu class.
========================================================================*/

eq.c.DlgContainer.addClass('Menu', function() {
   eq.c.DlgContainer.call(this);
});

/*------------------------------------------------------------------------
   Menu.setup()
   Setup Menu from DLG request.
------------------------------------------------------------------------*/

eq.c.Menu.prototype.setup = function(dlg, rq, et) {
   et.tg = 'LI';
   et.cla = [ 'eq-menu', 'eq-control' ];
   et.et = {
      tg : 'UL',
      ev  : { c : [ 'mousedown' ] }
   };
   eq.c.DlgContainer.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   Menu.setAttr()
   Set Menu attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.Menu.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.text: {
         var rt = et.rt = et.rt || [];
         rt.push({ t : tag, v : rq.decodeString() });
         break;
      }

      case rt.align: {
         var rt = et.rt = et.rt || [];
         rt.push({ t : tag, v : rq.decodeUint() });
         break;
      }

      default:
         eq.c.DlgContainer.prototype.setAttr.call(this, dlg, rq, tag, et);
   }
};

/** @preserve $Id: 10-poptext.js,v 28.8 2021/06/01 18:12:29 mike Exp $
*//*======================================================================
   class PopText extends DlgControl
   DLG PopText class.
========================================================================*/

eq.c.DlgControl.addClass('PopText', function() {
   eq.c.DlgControl.call(this);

   // Default height.
   this.defaultHeight = 8;
   var api = eq.api;
   this.defaultPxHeight = api.pxHeight(this.defaultHeight);
   this.defaultCssHeight = api.cssHeight(this.defaultHeight);
});

/*------------------------------------------------------------------------
   PopText.setup()
   Setup PopText from DLG request.
------------------------------------------------------------------------*/

eq.c.PopText.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-poptext', 'eq-control', 'eq-border' ];
   et.ev  = { c : [ 'input', 'mousedown' ] };
   et.et = { tg  : 'SPAN' };
   eq.c.DlgControl.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   PopText.setAttr()
   Set PopText attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.PopText.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.content: {
         var
            l = rq.decodeUint(),
            rt = et.rt = et.rt || [],
            v = [];
         v.length = l;
         for (var i = 0; i < l; i++)
            v[i] = rq.decodeString();
         rt.push({ t : tag, v : v });
         break;
      }

      case rt.activeLine: {
         var rt = et.rt = et.rt || [];
         rt.push({ t : tag, v : rq.decodeUint() });
         break;
      }

      case rt.boxSize: {
         var
            w = rq.decodeUint(),
            h = rq.decodeUint(),
            rt = et.rt = et.rt || [],
            cw, ch, mh;
         if (this.sizeRaster) {
            var api = eq.api;
            if (w)
               cw = api.cssWidth(w);
            if (h && h > this.defaultHeight)
               ch = api.cssHeight(h);
            else
               ch = this.defaultCssHeight;
         }
         else {
            if (w)
               cw = w + 'px';
            if (h && h > this.defaultPxHeight)
               ch = h + 'px';
            else
               ch = this.defaultCssHeight;
         }
         rt.push({ t : tag, w : cw, h : ch });
         break;
      }

      default:
         eq.c.DlgControl.prototype.setAttr.call(this, dlg, rq, tag, et);
         return;
   }
};

/*------------------------------------------------------------------------
   PopText.setRuleAttr()
   Set 'rule' attribute from DLG request.
------------------------------------------------------------------------*/

eq.c.PopText.prototype.setRuleAttr = function(dlg, rq, et, rule) {
   var ev = et.ev = et.ev || {};
   ev.a = rule ? [ 'input' ] : null;
};

/** @preserve $Id: 10-progressbar.js,v 28.5 2021/06/01 18:12:29 mike Exp $
*//*======================================================================
   class ProgressBar extends DlgControl
   DLG ProgressBar class.
========================================================================*/

eq.c.DlgControl.addClass('ProgressBar', function() {
   eq.c.DlgControl.call(this);
});

eq.c.ProgressBar.prototype.focusable = false;

/*------------------------------------------------------------------------
   ProgressBar.setup()
   Setup ProgressBar from DLG request.
------------------------------------------------------------------------*/

eq.c.ProgressBar.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-progressbar', 'eq-control', 'eq-border' ];
   eq.c.DlgControl.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   ProgressBar.setAttr()
   Set ProgressBar attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.ProgressBar.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.vertical: {
         var rt = et.rt = et.rt || [];
         rt.push({ t : tag, v : rq.decodeUint() });
         break;
      }

      case rt.value:
      case rt.text: {
         var rt = et.rt = et.rt || [];
         rt.push({ t : tag, v : rq.decodeString() });
         break;
      }

      default:
         eq.c.DlgControl.prototype.setAttr.call(this, dlg, rq, tag, et);
   }
};

/** @preserve $Id: 10-pushbutton.js,v 28.22 2021/07/16 15:24:34 mike Exp $
*//*======================================================================
   class PushButton extends DlgControl
   DLG PushButton class.
========================================================================*/

eq.c.DlgControl.addClass('PushButton', function() {
   eq.c.DlgControl.call(this);
});

eq.c.PushButton.prototype.textPos = 0;

/*------------------------------------------------------------------------
   PushButton.setup()
   Setup PushButton from DLG request.
------------------------------------------------------------------------*/

eq.c.PushButton.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-pushbutton', 'eq-control', 'eq-border' ];
   et.et = {
      tg : 'BUTTON',
      at : { 'type' : 'button' },
      ev : {
         // Element is focusable and can submit,
         // use 'mousedown', not 'click'.
         a : [ 'mousedown' ]
      }
   };
   eq.c.DlgControl.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   PushButton.setAttr()
   Set PushButton attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.PushButton.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.textPos:
      case rt.text:
      case rt.icon:
      case rt.iconActive:
         eq.c.DlgControl.isSetup(rq, et, 2);
         break;

      default:
         eq.c.DlgControl.prototype.setAttr.call(this, dlg, rq, tag, et);
         return;
   }

   switch (tag) {
      case rt.textPos: {
         var clr = et.clr = et.clr || [];
         if ((this.textPos = rq.decodeUint()) === 0) {
            // no icon
            clr.push('eq-horz');
            clr.push('eq-vert');
            delete this.minHeight;
         }
         else {
            // icon
            var ch, cla = et.cla = et.cla || [];
            ch = et.et.ch = [ { tg : 'DIV' } ];
            ch = ch[0].ch = [{},{}];
            switch (this.textPos) {
               case 1: { // right
                  ch[0].tg = 'IMG';
                  ch[1].tg = 'SPAN';
                  clr.push('eq-vert');
                  cla.push('eq-horz');
                  break;
               }
               case 2: { // bottom
                  ch[0].tg = 'IMG';
                  ch[1].tg = 'SPAN';
                  clr.push('eq-horz');
                  cla.push('eq-vert');
                  break;
               }
               case 3: { // left
                  ch[0].tg = 'SPAN';
                  ch[1].tg = 'IMG';
                  clr.push('eq-vert');
                  cla.push('eq-horz');
                  break;
               }
               case 4: { // top
                  ch[0].tg = 'SPAN';
                  ch[1].tg = 'IMG';
                  clr.push('eq-horz');
                  cla.push('eq-vert');
                  break;
               }
               default:
                  throw new Error("PushButton.setAttr(" + this.id +
                     ") invalid textPos " + this.textPos);
            }
            this.minHeight = false;
         }
         break;
      }

      case rt.text: {
         var s = rq.decodeString();
         if (this.textPos === 0) {
            // no icon
            et = et.et;
         }
         else {
            // icon
            var ch = et.et.ch = et.et.ch || [ { tg : 'DIV' } ];
            et = ch[0];
            ch = et.ch = et.ch || [{},{}];
            switch (this.textPos) {
               case 1: // right
               case 2: // bottom
                  et = ch[1];
                  break;
               default:
                  et = ch[0];
            }
         }
         et.hca = s;
         break;
      }

      case rt.icon: {
         var s = eq.api.makeUrl(rq.decodeString());
         if (this.textPos) {
            var ch = et.et.ch = et.et.ch || [ { tg : 'DIV' } ];
            et = ch[0];
            ch = et.ch = et.ch || [{},{}];
            switch (this.textPos) {
               case 1: // right
               case 2: // bottom
                  et = ch[0];
                  break;
               default:
                  et = ch[1];
            }
            et.at = et.at || {};
            et.at['src'] = s.length ? s : null;
         }
         break;
      }

      case rt.iconActive:
         // TODO
         eq.api.makeUrl(rq.decodeString());
         break;
   }
};

/*------------------------------------------------------------------------
   PushButton.setSensitive()
   Set element-specific 'sensitive' properties.
------------------------------------------------------------------------*/

eq.c.PushButton.prototype.setSensitive = function(dlg, rq, et) {
   eq.c.DlgControl.isSetup(rq, et, 2);
   var at = et.et.at = et.et.at || {};
   at.disabled = this.sensitive ? null : true;
};

/** @preserve $Id: 10-radiobutton.js,v 28.15 2021/07/13 13:18:15 mike Exp $
*//*======================================================================
   class RadioButton extends DlgControl
   DLG RadioButton class.
========================================================================*/

eq.c.DlgControl.addClass('RadioButton', function() {
   eq.c.DlgControl.call(this);
});

/*------------------------------------------------------------------------
   RadioButton.setup()
   Setup RadioButton from DLG request.
------------------------------------------------------------------------*/

eq.c.RadioButton.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-radiobutton', 'eq-control' ];
   et.et = {
      ch : [
         { id : '*' },
         { id : '*' }
      ]
   };
   eq.c.DlgControl.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   RadioButton.setAttr()
   Set RadioButton attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.RadioButton.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.textPos:
      case rt.radioGroup:
      case rt.active:
      case rt.text:
         if (eq.c.DlgControl.isSetup(rq, et, 2))
            et.et.ch = et.et.ch || [{},{}];
         break;

      default:
         eq.c.DlgControl.prototype.setAttr.call(this, dlg, rq, tag, et);
         return;
   }

   switch (tag) {
      case rt.textPos: {
         // Button index, 0 = label right, 1 = label left.
         this.iBtn = rq.decodeUint();
         // Button.
         var ch = et.et.ch[this.iBtn];
         ch.tg = 'INPUT';
         ch.at = { 'type' : 'radio' };
         ch.ev = {
            c : [ 'click' ]
         };
         // Label.
         ch = et.et.ch[this.iBtn ? 0 : 1];
         ch.tg = 'LABEL';
         ch.at = { 'for' : et.id + '-' + (this.iBtn + 1) };
         break;
      }

      case rt.radioGroup: {
         var ch = et.et.ch[this.iBtn];
         ch.at = ch.at || {};
         ch.at.name = rq.decodeString();
         break;
      }

      case rt.active: {
         var ch = et.et.ch[this.iBtn];
         ch.at = ch.at || {};
         ch.at.checked = rq.decodeUint() !== 0;
         break;
      }

      case rt.text: {
         var ch = et.et.ch[this.iBtn ? 0 : 1];
         ch.hca = rq.decodeString();
         break;
      }
   }
};

/*------------------------------------------------------------------------
   RadioButton.setRuleAttr()
   Set 'rule' attribute from DLG request.
------------------------------------------------------------------------*/

eq.c.RadioButton.prototype.setRuleAttr = function(dlg, rq, et, rule) {
   if (eq.c.DlgControl.isSetup(rq, et, 2))
      et.et.ch = et.et.ch || [{},{}];
   var ch = et.et.ch[this.iBtn], ev = ch.ev = ch.ev || {};
   // Element is focusable and can submit,
   // use 'mousedown', not 'click'.
   ev.a = rule ? [ 'mousedown' ] : null;
};

/*------------------------------------------------------------------------
   RadioButton.setSensitive()
   Set element-specific 'sensitive' properties.
------------------------------------------------------------------------*/

eq.c.RadioButton.prototype.setSensitive = function(dlg, rq, et) {
   if (eq.c.DlgControl.isSetup(rq, et, 2))
      et.et.ch = et.et.ch || [{},{}];
   var ch = et.et.ch[this.iBtn], at = ch.at = ch.at || {};
   at.disabled = this.sensitive ? null : true;
};

/** @preserve $Id: 10-separator.js,v 28.4 2021/05/28 15:55:16 mike Exp $
*//*======================================================================
   class Separator extends DlgControl
   DLG Separator class.
========================================================================*/

eq.c.DlgControl.addClass('Separator', function() {
   eq.c.DlgControl.call(this);
});

eq.c.Separator.prototype.minWidth = true;
eq.c.Separator.prototype.focusable = false;

/*------------------------------------------------------------------------
   Separator.setup()
   Setup Separator from DLG request.
------------------------------------------------------------------------*/

eq.c.Separator.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-separator', 'eq-control' ];
   eq.c.DlgControl.prototype.setup.call(this, dlg, rq, et);
};

/** @preserve $Id: 10-splitter.js,v 28.5 2021/06/01 18:12:29 mike Exp $
*//*======================================================================
   class Splitter extends DlgContainer
   DLG Splitter class.
========================================================================*/

eq.c.DlgContainer.addClass('Splitter', function() {
   eq.c.DlgContainer.call(this);
});

eq.c.Splitter.prototype.vertical = false;

/*------------------------------------------------------------------------
   Splitter.setup()
   Setup Splitter from DLG request.
------------------------------------------------------------------------*/

eq.c.Splitter.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-splitter', 'eq-control', 'eq-border' ];
   et.ev = { c : [ 'mousedown' ] };
   eq.c.DlgContainer.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   Splitter.setAttr()
   Set Splitter attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.Splitter.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.panes: {
         var rt = et.rt = et.rt || [];
         rt.push({ t : tag });
         break;
      }

      case rt.vertical: {
         var cl;
         if (rq.decodeUint()) {
            this.vertical = true;
            cl = et.cla = et.cla || [];
         }
         else {
            this.vertical = false;
            cl = et.clr = et.clr || [];
         }
         cl.push('eq-vert');
         break;
      }

      case rt.position: {
         var
            iv = rq.decodeUint(),
            rt = et.rt = et.rt || [];
         if (this.sizeRaster)
            iv = this.vertical ? eq.api.pxHeight(iv) : eq.api.pxWidth(iv);
         rt.push({ t : tag, v : iv });
         break;
      }

      case rt.quickExpand: {
         var cl;
         if (rq.decodeUint())
            cl = et.cla = et.cla || [];
         else
            cl = et.clr = et.clr || [];
         cl.push('eq-button');
         break;
      }

      default:
         eq.c.DlgContainer.prototype.setAttr.call(this, dlg, rq, tag, et);
   }
};

/*------------------------------------------------------------------------
   Splitter.setRuleAttr()
   Set 'rule' attribute from DLG request.
------------------------------------------------------------------------*/

eq.c.Splitter.prototype.setRuleAttr = function(dlg, rq, et, rule) {
   var cl;
   if (rule)
      cl = et.cla = et.cla || [];
   else
      cl = et.clr = et.clr || [];
   cl.push('eq-rule');
};

/** @preserve $Id: 10-statictext.js,v 28.16 2021/05/07 15:05:36 mike Exp $
*//*======================================================================
   class StaticText extends DlgControl
   DLG StaticText class.
========================================================================*/

eq.c.DlgControl.addClass('StaticText', function() {
   eq.c.DlgControl.call(this);
});

eq.c.StaticText.prototype.focusable = false;

/*------------------------------------------------------------------------
   StaticText.setup()
   Setup StaticText from DLG request.
------------------------------------------------------------------------*/

eq.c.StaticText.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-statictext', 'eq-control' ];
   et.et = {
      tg : 'SPAN'
   };
   eq.c.DlgControl.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   StaticText.setAttr()
   Set StaticText attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.StaticText.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.text:
      case rt.align:
         eq.c.DlgControl.isSetup(rq, et, 2);
         break;

      default:
         eq.c.DlgControl.prototype.setAttr.call(this, dlg, rq, tag, et);
         return;
   }

   switch (tag) {
      case rt.text:
         et.et.hca = rq.decodeString();
         break;

      case rt.align: {
         var st = et.et.st = et.et.st || {};
         switch (rq.decodeUint()) {
            case 0:
               st['text-align'] = null;
               break;
            case 1:
               st['text-align'] = 'center';
               break;
            case 2:
               st['text-align'] = 'right';
         }
         break;
      }
   }
};

/** @preserve $Id: 10-statusbar.js,v 28.6 2021/05/28 15:56:12 mike Exp $
*//*======================================================================
   class StatusBar extends DlgContainer
   DLG StatusBar class.
========================================================================*/

eq.c.DlgContainer.addClass('StatusBar', function() {
   eq.c.DlgContainer.call(this);
});

/*------------------------------------------------------------------------
   StatusBar.setup()
   Setup StatusBar from DLG request.
------------------------------------------------------------------------*/

eq.c.StatusBar.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-statusbar', 'eq-container' ];
   eq.c.DlgContainer.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   StatusBar.setAttr()
   Set StatusBar attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.StatusBar.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.posSize: {
         var
            st = et.st = et.st || {},
            iv = rq.decodeUint();
            this.sizeRaster = !!(iv & 1);
            if (iv = rq.decodeInt()) { // height
               if (this.sizeRaster)
                  st['height'] = eq.api.cssHeight(iv);
               else
                  st['height'] = iv + 'px';
            }
            else
               st['height'] = eq.api.cssHeight(1);
         break;
      }

      default:
         eq.c.DlgContainer.prototype.setAttr.call(this, dlg, rq, tag, et);
   }
};

/** @preserve $Id: 10-tabbox.js,v 28.9 2021/07/16 16:20:00 mike Exp $
*//*======================================================================
   class TabBox extends DlgContainer
   DLG TabBox class.
========================================================================*/

eq.c.DlgContainer.addClass('TabBox', function() {
   eq.c.DlgContainer.call(this);
});

eq.c.TabBox.prototype.focusable = true;

/*------------------------------------------------------------------------
   TabBox.setup()
   Setup TabBox from DLG request.
------------------------------------------------------------------------*/

eq.c.TabBox.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-tabbox', 'eq-control', 'eq-border' ];
   et.et = {
      cla : [ 'eq-container' ]
   };
   eq.c.DlgContainer.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   TabBox.setAttr()
   Set TabBox attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.TabBox.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.tabs: {
         var
            rt = et.rt = et.rt || [],
            v = [], s;
         while (s = rq.decodeString())
            v.push(s);
         rt.push({ t : tag, v : v });
         break;
      }

      case rt.selectedTab: {
         var rt = et.rt = et.rt || [];
         rt.push({ t : tag, v : rq.decodeString() });
         break;
      }

      default:
         eq.c.DlgContainer.prototype.setAttr.call(this, dlg, rq, tag, et);
   }
};

/*------------------------------------------------------------------------
   TabBox.setRuleAttr()
   Set 'rule' attribute from DLG request.
------------------------------------------------------------------------*/

eq.c.TabBox.prototype.setRuleAttr = function(dlg, rq, et, rule) {
   var cl;
   if (rule)
      cl = et.cla = et.cla || [];
   else
      cl = et.clr = et.clr || [];
   cl.push('eq-rule');
};

/*------------------------------------------------------------------------
   TabBox.updateTabOrder()
   Act upon main thread requesting to update tab order.
------------------------------------------------------------------------*/

eq.c.TabBox.prototype.onUpdateTabOrder = function(arg) {
   for (var i = 0, l = this.ch.length; i < l; i++) {
      var ch = this.ch[i];
      if (ch.visible && ch.sensitive && arg-- === 0)
         delete ch.inactive;
      else
         ch.inactive = true;
   }
};

/** @preserve $Id: 10-toolbar.js,v 28.4 2021/05/28 15:55:16 mike Exp $
*//*======================================================================
   class ToolBar extends DlgContainer
   DLG ToolBar class.
========================================================================*/

eq.c.DlgContainer.addClass('ToolBar', function() {
   eq.c.DlgContainer.call(this);
});

/*------------------------------------------------------------------------
   ToolBar.setup()
   Setup ToolBar from DLG request.
------------------------------------------------------------------------*/

eq.c.ToolBar.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-toolbar', 'eq-container' ];
   eq.c.DlgContainer.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   ToolBar.setAttr()
   Set ToolBar attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.ToolBar.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.posSize: {
         var
            cla = et.cla = et.cla || [],
            clr = et.clr = et.clr || [];
         switch (rq.decodeUint()) {
            case 1:
               cla.push('eq-bottom');
               clr.push('eq-top');
               clr.push('eq-left');
               clr.push('eq-right');
               break;
            case 2:
               cla.push('eq-left');
               clr.push('eq-top');
               clr.push('eq-bottom');
               clr.push('eq-right');
               break;
            case 3:
               cla.push('eq-right');
               clr.push('eq-top');
               clr.push('eq-bottom');
               clr.push('eq-left');
               break;
            default:
               cla.push('eq-top');
               clr.push('eq-bottom');
               clr.push('eq-left');
               clr.push('eq-right');
         }
         break;
      }

      default:
         eq.c.DlgContainer.prototype.setAttr.call(this, dlg, rq, tag, et);
   }
};

/*------------------------------------------------------------------------
   ToolBar.containerFocusable()
   Check 'focusable' based on container.
------------------------------------------------------------------------*/

eq.c.ToolBar.prototype.containerFocusable = function() {
   // Objects in ToolBar not focusable.
   return false;
};

/*------------------------------------------------------------------------
   ToolBar.updateTabOrder()
   Fill current tab order array.
------------------------------------------------------------------------*/

eq.c.ToolBar.prototype.updateTabOrder = function(tabOrder) {
   // Objects in ToolBar not focusable.
}

/** @preserve $Id: 10-tree.js,v 28.5 2021/07/04 16:13:09 mike Exp $
*//*======================================================================
   class Tree extends DlgControl
   DLG Tree class.
========================================================================*/

eq.c.DlgControl.addClass('Tree', function() {
   eq.c.DlgControl.call(this);
});

/*------------------------------------------------------------------------
   Tree.setup()
   Setup Tree from DLG request.
------------------------------------------------------------------------*/

eq.c.Tree.prototype.setup = function(dlg, rq, et) {
   et.cla = [ 'eq-tree', 'eq-control', 'eq-border' ];
   et.ev = { c : [ 'mousedown', 'change', 'status' ] };
   et.et = { cla : [ 'eq-pane' ] };
   eq.c.DlgControl.prototype.setup.call(this, dlg, rq, et);
};

/*------------------------------------------------------------------------
   Tree.setAttr()
   Set Tree attributes from DLG request.
------------------------------------------------------------------------*/

eq.c.Tree.prototype.setAttr = function(dlg, rq, tag, et) {
   var rt = eq.RqTag;
   switch (tag) {
      case rt.content: {
         var
            rt = et.rt = et.rt || [],
            v = [];
         this.setContent(rq, v);
         rt.push({ t : tag, v : v });
         break;
      }

      case rt.activeLine: {
         var rt = et.rt = et.rt || [];
         rt.push({ t : tag, v : rq.decodeUint() });
         break;
      }

      case rt.status: {
         var
            l = rq.decodeUint(),
            rt = et.rt = et.rt || [],
            v = [];
         v.length = l;
         for (var i = 0; i < l; i++)
            v[i] = rq.decodeUint();
         rt.push({ t : tag, v : v });
         break;
      }

      case rt.ruleMode:
         this.ruleMode = rq.decodeUint();
         break;

      default:
         eq.c.DlgControl.prototype.setAttr.call(this, dlg, rq, tag, et);
   }
};

/*------------------------------------------------------------------------
   Tree.setContent()
   Set Tree content from DLG request.
------------------------------------------------------------------------*/

eq.c.Tree.prototype.setContent = function(rq, v) {
   // Number of items.
   var l = rq.decodeUint();

   // Items.
   v.length = l;
   for (var i = 0; i < l; i++)
   {
      var
         f = rq.decodeUint(),    // Flags
         hc = rq.decodeString(); // Text or HTML content
      if (f === 0) {
         // Leaf.
         v[i] = hc;
      }
      else {
         // Node, children.
         var ch = [];
         this.setContent(rq, ch, hc);
         v[i] = {
            f : f,
            hc : hc,
            ch : ch
         };
      }
   }
};

/*------------------------------------------------------------------------
   Tree.setRuleAttr()
   Set 'rule' attribute from DLG request.
------------------------------------------------------------------------*/

eq.c.Tree.prototype.setRuleAttr = function(dlg, rq, et, rule) {
   var ev = et.ev = et.ev || {};
   if (rule) {
      // Element is focusable and can submit,
      // use 'mousedown', not 'click'.
      ev.a = [ 'mousedown', 'keydown' ];
      switch (this.ruleMode) {
         case 1:
            ev.m = 1;
            ev.f = null;
            break;
         case 2:
            ev.m = 2;
            ev.f = 1;
            break;
         case 3:
            ev.m = 1;
            ev.f = 1;
            break;
         default:
            ev.m = 2;
            ev.f = null;
      }
   }
   else {
      ev.a = null;
      ev.m = null;
      ev.f = null;
   }
};

/** @preserve $Id: 10-windowmenu.js,v 28.3 2021/03/11 10:01:28 mike Exp $
*//*======================================================================
   class WindowMenu extends DlgControl
   DLG WindowMenu class.
========================================================================*/

eq.c.DlgControl.addClass('WindowMenu', function() {
   eq.c.DlgControl.call(this);
});

/** @preserve $Id: 99-worker.js,v 28.32 2021/07/16 16:19:43 mike Exp $
*//*======================================================================
   class Api extends Obj
   Application API protocol thread.
========================================================================*/

eq.c.Obj.subClass('Api', function() {
   eq.c.Obj.call(this);
   this.reset();

   // Base URL.
   this.baseUrl = null;

   // Dialog map.
   this.dlg = null;

   // Reconnect timeout.
   this.reconnectTimeout = 60; // Seconds
});

/*------------------------------------------------------------------------
   Api.reset()
   Reset API protocol thread.
------------------------------------------------------------------------*/

eq.c.Api.prototype.reset = function() {
   this.dlgRq = new eq.c.DlgRq(); // DLG request
   this.dlgRs = new eq.c.DlgRs(); // DLG response
   this.pendingCall = false;
   this.currDlg = undefined;
};

/*------------------------------------------------------------------------
   Api.resetDlg()
   Reset DLG session.
------------------------------------------------------------------------*/

eq.c.Api.prototype.resetDlg = function() {
   if (this.dlg) {
      this.dlg.forEach(eq.c.Api.cbCloseDlg);
      this.dlg.clear();
      this.dlg = null;
   }
};

eq.c.Api.cbCloseDlg = function(d) {
   d.close();
};

/*------------------------------------------------------------------------
   Api.invalidateDlg()
   Invalidate DLG session, invoked from webdReceived(DlgRq.mode.add).
------------------------------------------------------------------------*/

eq.c.Api.prototype.invalidateDlg = function(rq) {
   var dlg = this.dlg, d = [];
   dlg.forEach(eq.c.Api.cbInvalidateDlg, { d : d });
   dlg.clear();
   if (d.length) {
      rq.msg.push({
         o : eq.MainOp.invalidate,
         d : d
      });
   }
};

eq.c.Api.cbInvalidateDlg = function(d) {
   this.d.push(d.id);
   d.close();
};

/*------------------------------------------------------------------------
   Api.gridWidth() .gridHeight()
   Return width/height in grid units.
------------------------------------------------------------------------*/

eq.c.Api.prototype.gridWidth = function(w) {
   if (w && this.grid.wp)
      return Math.round(w / this.grid.wp);
   return 0;
}

eq.c.Api.prototype.gridHeight = function(h) {
   if (h && this.grid.hp)
      return Math.round(h / this.grid.hp);
   return 0;
}

/*------------------------------------------------------------------------
   Api.pxWidth() .pxHeight()
   Return width/height in pixel units.
------------------------------------------------------------------------*/

eq.c.Api.prototype.pxWidth = function(w) {
   if (w)
      return Math.round(w * this.grid.wp * 1000) / 1000;
   return 0;
}

eq.c.Api.prototype.pxHeight = function(h) {
   if (h)
      return Math.round(h * this.grid.hp * 1000) / 1000;
   return 0;
}

/*------------------------------------------------------------------------
   Api.cssWidth() .cssHeight()
   Return width/height in CSS units.
------------------------------------------------------------------------*/

eq.c.Api.prototype.cssWidth = function(w) {
   if (w)
      return (Math.round(w * this.grid.wf * 1000) / 1000) + 'ch';
   return '0';
}

eq.c.Api.prototype.cssHeight = function(h) {
   if (h)
      return (Math.round(h * this.grid.hf * 1000) / 1000) + 'em';
   return '0';
}

/*------------------------------------------------------------------------
   Api.makeUrl()
   Prepend base URL if necessary.
------------------------------------------------------------------------*/

eq.c.Api.prototype.makeUrl = function(u) {
   return u.length && this.baseUrl !== null
                   && !eq.c.Api.isAbsoluteUrl.test(u)
        ? this.baseUrl + u : u;
}

eq.c.Api.isAbsoluteUrl = new RegExp('^(/|[A-Z0-9_]+:)', 'i');

/*------------------------------------------------------------------------
   static Api.mainReceived()
   Process message from main thread.
------------------------------------------------------------------------*/

eq.c.Api.mainReceived = function(m) {
   var self = eq.api, op = eq.ApiOp;
   switch (m.data.o) {
      case op.documentComplete:
         // Document is complete.
         self.documentComplete = true;
         // Save layout grid.
         self.grid = m.data.d;
         // Start or resume application if already possible.
         self.startOrResumeApplication();
         break;

      case op.openWebSocket:
         // Open WebSocket connection.
         self.app = m.data.app;
         self.webdURL = m.data.url;
         self.webdOpen();
         break;

      case op.onEvent: {
         if (!self.pendingCall)
            throw new Error("mainReceived onEvent: " +
               "invalid, no pending call");
         self.pendingCall = false;

         if (self.currDlg === undefined)
            throw new Error("mainReceived onEvent: " +
               "invalid, no current DLG");
         if (!self.currDlg.root)
            throw new Error("mainReceived onEvent: " +
               "invalid, no DLG root object");

         // Submit event to application.
         var rs = self.dlgRs;
         try {
            rs.encodeUint(eq.c.DlgRs.op.event);

            // 1st and 2nd event, Element id.
            var ev = m.data.d, id = ev.id, did;
            if (!id)
               id = did !== undefined ? did : (did = self.currDlg.root.id);
            rs.encodeString(id);
            if (ev.ev !== undefined) {
               if (!(id = ev.ev.id))
                  id = did !== undefined ? did : (did = self.currDlg.root.id);
               rs.encodeString(id);
            }
            else
               rs.encodeUint(0); // No 2nd event.

            // Changed elements.
            var pack = eq.c.Api.packValue;
            for (var i = 0, l = ev.changed.length; i < l; i++) {
               var el = ev.changed[i];
               if (!(id = el.id))
                  id = did !== undefined ? did : (did = self.currDlg.root.id);
               rs.encodeString(id);
               if (el.v === undefined)
                  pack(rs, id, el);  // Single value.
               else {
                  // Multiple values.
                  for (var vi = 0, vl = el.v.length; vi < vl; vi++)
                     pack(rs, id, el.v[vi]);
               }
               rs.encodeUint(0); // End of value(s).
            }
            rs.encodeUint(0); // End of changed elements.

            rs.encodeUint(0); // End of message.
            self.ws.send(rs.buffer());
            self.currDlg = undefined;
         }
         catch (e) {
            self.currDlg = undefined;
            rs.dispose();
            throw e;
         }
         break;
      }

      case op.onPopupEvent: {
         if (!self.pendingCall)
            throw new Error("mainReceived onPopupEvent: " +
               "invalid, no pending call");
         self.pendingCall = false;

         // Submit event to application.
         var rs = self.dlgRs;
         try {
            rs.encodeUint(eq.c.DlgRs.op.popup);

            // Button number.
            rs.encodeInt(m.data.d);

            rs.encodeUint(0); // End of message.
            self.ws.send(rs.buffer());
         }
         catch (e) {
            rs.dispose();
            throw e;
         }
         break;
      }

      case op.onDraw: {
         if (!self.pendingCall)
            throw new Error("mainReceived onDraw: " +
               "invalid, no pending call");
         self.pendingCall = false;

         // Submit event to application.
         var rs = self.dlgRs;
         try {
            rs.encodeUint(eq.c.DlgRs.op.draw);
            rs.encodeUint(0); // End of message.
            self.ws.send(rs.buffer());
         }
         catch (e) {
            rs.dispose();
            throw e;
         }
         break;
      }

      case op.invalidate: {
         var id = m.data.d, dlg = self.dlg.get(id);
         if (dlg === undefined)
            throw new Error("mainReceived invalidate: DLG " +
               id + " does not exist");
         dlg.close();
         self.dlg.delete(id);

         // Notify application.
         var rs = self.dlgRs;
         try {
            rs.encodeUint(eq.c.DlgRs.op.invalidate);
            rs.encodeUint(id); // Dialog id.

            rs.encodeUint(0); // End of message.
            self.ws.send(rs.buffer());
         }
         catch (e) {
            rs.dispose();
            throw e;
         }
         break;
      }

      case op.updateTabOrder: {
         var id = m.data.d, dlg = self.dlg.get(id), a = m.data.a;
         if (dlg === undefined)
            throw new Error("mainReceived updateTabOrder: DLG " +
               id + " does not exist");
         if (a !== undefined) {
            var ct = dlg.findById(a.id);
            ct.onUpdateTabOrder(a.arg);
         }
         dlg.updateTabOrder();
         // Notify main thread.
         postMessage([{
            o : eq.MainOp.tabOrderUpdated,
            d : dlg.tabOrder
         }]);
         break;
      }

      default:
         throw new Error("mainReceived invalid op: " + m.data.o);
   }
};

eq.c.Api.packValue = function(rs, id, v) {
   var vt = eq.RsValType;
   rs.encodeUint(v.ty);
   switch (v.ty) {
      case vt.idFocus:
      case vt.textValue:
         // String value.
         if (v.sv === undefined)
            throw new Error("mainReceived onEvent: RsValType " + v.ty +
               " sv expected");
         rs.encodeString(v.sv);
         break;

      case vt.ruleOnce:
         // INT value.
         if (v.iv === undefined)
            throw new Error("mainReceived onEvent: RsValType " + v.ty +
               " iv expected");
         rs.encodeInt(v.iv);
         break;

      case vt.buttonChecked:
      case vt.caretPos:
      case vt.activeLine:
      case vt.topItem:
      case vt.selectedTab:
      case vt.ruleKey:
      case vt.mouseBtns:
      case vt.mouseCnt:
      case vt.mouseMods:
         // UINT value.
         if (v.iv === undefined)
            throw new Error("mainReceived onEvent: RsValType " + v.ty +
               " iv expected");
         rs.encodeUint(v.iv);
         break;

      case vt.splitterPos: {
         // UINT position, grid-adjusted.
         var ct = eq.api.currDlg.findById(id), iv = v.iv;
         if (iv === undefined)
            throw new Error("mainReceived onEvent: RsValType " + v.ty +
               " iv expected");
         if (ct.sizeRaster)
            iv = ct.vertical ? eq.api.gridHeight(iv) : eq.api.gridWidth(iv);
         rs.encodeUint(iv);
         break;
      }

      case vt.status: {
         // ARRAY value.
         var vv = v.vv, l;
         if (vv === undefined)
            throw new Error("mainReceived onEvent: RsValType " + v.ty +
               " vv expected");
         rs.encodeUint(l = vv.length);
         for (var i = 0; i < l; i++)
            rs.encodeUint(vv[i]);
         break;
      }

      default:
         throw new Error("mainReceived onEvent: unknown RsValType " + v.ty);
   }
};

/*------------------------------------------------------------------------
   Api.startOrResumeApplication()
   Start or resume application.
------------------------------------------------------------------------*/

eq.c.Api.prototype.startOrResumeApplication = function() {
   if (this.webdOpened && this.documentComplete) {
      if (this.startSession) {
         this.startSession = undefined;
         if (this.dlg !== null)
            throw new Error("BUG: startApplication unexpected DLG session");
         // Initialize DLG session, start application.
         this.dlg = new Map();
         var rs = this.dlgRs;
         try {
            rs.encodeUint(eq.c.DlgRs.op.start);

            // Send user interface class names.
            for (var i = 0, l = eq.d.length; i < l; i++)
               rs.encodeString(eq.d[i].name);
            rs.encodeUint(0);

            rs.encodeUint(0); // End of message.
            this.ws.send(rs.buffer());
         }
         catch (e) {
            rs.dispose();
            throw e;
         }
      }
      else if (this.resumeSession) {
         if (this.dlg === null)
            this.dlg = new Map();
         // Resume application.
         var rs = this.dlgRs;
         try {
            rs.encodeUint(eq.c.DlgRs.op.resume);
            rs.encodeUint(0); // End of message.
            this.ws.send(rs.buffer());
         }
         catch (e) {
            rs.dispose();
            throw e;
         }
      }
   }
};

/*------------------------------------------------------------------------
   Api.webdOpen()
   Open WebSocket connection to webd server.
------------------------------------------------------------------------*/

eq.c.Api.prototype.webdOpen = function() {
   if (self.webdHandshakeTimer !== undefined)
      throw new Error( "BUG: webdOpen pending handshake");
   this.ws = new WebSocket(this.webdURL + this.app,
      'eloquence.webdlg.' + eq.p.version);
   this.ws.binaryType = 'arraybuffer';
   this.ws.onopen = eq.c.Api.webdOpened;
   this.ws.onmessage = eq.c.Api.webdReceived;
   this.ws.onclose = eq.c.Api.webdClosed;
   this.ws.onerror = eq.c.Api.webdFailed;
};

/*------------------------------------------------------------------------
   static Api.webdReceived()
   Process WebSocket message from webd server.
------------------------------------------------------------------------*/

eq.c.Api.webdReceived = function(m) {
   var self = eq.api, rq = self.dlgRq, md = eq.c.DlgRq.mode;
   try {
      rq.assign(m.data);

      while (true) {
         if (!(rq.mode = rq.decodeUint()))
            break;

         if (self.currDlg !== undefined)
            throw new Error("webdReceived: invalid request " + rq.mode +
               " during DLG call");
         if (self.pendingCall)
            throw new Error("webdReceived: invalid request " + rq.mode +
               " while call pending");
         if (self.resumeSession !== undefined)
            switch (rq.mode) {
               case md.global:
               case md.add:
               case md.POPUP_BOX:
                  break;
               default:
                  throw new Error("webdReceived: invalid request " + rq.mode +
                     " while resuming session");
            }

         switch (rq.mode) {
            case md.ssnNew:
            case md.ssnResume: {
               eq.c.Api.webdStopHandshakeTimer();
               self.sid = rq.decodeUint();
               var app = rq.decodeString();
               self.reconnectTimeout = rq.decodeUint();
               if (self.app === undefined)
                  throw new Error("webdReceived: invalid request " + rq.mode +
                     ": app undefined, expected: '" + app + "'");
               if (self.app !== app)
                  throw new Error("webdReceived: invalid request " + rq.mode +
                     ": expected app: '" + self.app +
                     "', received: '" + app + "'");
               if (self.startSession !== undefined)
                  throw new Error("webdReceived: invalid request " + rq.mode +
                     ": startSession unexpected, app: '" + app + "'");
               if (self.resumeSession !== undefined)
                  throw new Error("webdReceived: invalid request " + rq.mode +
                     ": resumeSession unexpected, app: '" + app + "'");
               if (rq.mode === md.ssnNew)
                  self.startSession = true;
               else
                  self.resumeSession = true;
               // Start or resume application if already possible.
               self.startOrResumeApplication();
               continue;
            }
         }

         if (self.dlg === null)
            throw new Error("webdReceived: invalid request " + rq.mode +
               ": DLG session closed");

         switch (rq.mode) {
            case md.global: {
               var tag, rt = eq.RqTag;
               while (tag = rq.decodeUint()) {
                  switch (tag) {
                     case rt.baseUrl: {
                        var url = rq.decodeString();
                        self.baseUrl = url.length ? url : null;
                        break;
                     }

                     case rt.helpBaseUrl:
                        // Forward to main thread.
                        rq.msg.push({
                           o : eq.MainOp.global,
                           d : { t : tag, v : rq.decodeString() }
                        });
                        break;

                     default:
                        throw new Error(
                           "webdReceived global: invalid tag " + tag);
                  }
               }
               break;
            }

            case md.add: {
               // Resume session?
               if (self.resumeSession) {
                  self.resumeSession = undefined;
                  self.invalidateDlg(rq);
                  rq.msg.push({ o : eq.MainOp.resume });
               }
               var
                  id = rq.decodeUint(),       // Dialog id.
                  dlg = self.dlg.get(id),
                  parent = rq.decodeString(); // Parent id (root if empty).
               if (parent) {
                  if (dlg === undefined)
                     throw new Error("webdReceived: DLG " + id +
                        " does not exist, mode: " + rq.mode);
                  parent = dlg.findById(parent);
               }
               else if (dlg === undefined)
                  self.dlg.set(id, (dlg = new eq.c.Dlg(id)));
               // Create control(s).
               dlg.addChild(rq, parent);
               // Invalidate tab order.
               dlg.tabOrderValid = false;
               break;
            }

            case md.modify: {
               var
                  id = rq.decodeUint(), // Dialog id.
                  dlg = self.dlg.get(id),
                  ctid = rq.decodeString(), // Control id.
                  ct, et;
               if (dlg === undefined)
                  throw new Error("webdReceived: DLG " + id +
                     " does not exist, mode: " + rq.mode);
               ct = dlg.findById(ctid);
               et = { id : ct.id };
               rq.msg.push({
                  o : eq.MainOp.modifyControl,
                  i : id,
                  d : et
               });
               ct.modify(dlg, rq, et);
               break;
            }

            case md.del: {
               var
                  id = rq.decodeUint(), // Dialog id.
                  dlg = self.dlg.get(id),
                  ctid = rq.decodeString(), // Control id.
                  ct;
               if (dlg === undefined)
                  throw new Error("webdReceived: DLG " + id +
                     " does not exist, mode: " + rq.mode);
               ct = dlg.findById(ctid);
               rq.msg.push({
                  o : eq.MainOp.deleteControl,
                  i : id,
                  d : ct.id
               });
               if (ct.del()) {
                  dlg.root = null;
                  dlg.close();
                  self.dlg.delete(id);
               }
               // Invalidate tab order.
               dlg.tabOrderValid = false;
               break;
            }

            case md.invalidate: {
               var id, dlg, d = [];
               while (id = rq.decodeUint()) { // Dialog id.
                  if ((dlg = self.dlg.get(id)) === undefined)
                     continue;
                  d.push(id);
                  dlg.close();
                  self.dlg.delete(id);
               }
               if (d.length) {
                  rq.msg.push({
                     o : eq.MainOp.invalidate,
                     d : d
                  });
               }
               break;
            }

            case md.DLG_DRAW: {
               var
                  id = rq.decodeUint(), // Dialog id.
                  dlg = self.dlg.get(id);
               if (dlg === undefined)
                  throw new Error("webdReceived: DLG " + id +
                     " does not exist, mode: " + rq.mode);
               self.pendingCall = true;
               rq.msg.push({
                  o : eq.MainOp.dlg_DRAW,
                  i : id
               });
               break;
            }

            case md.DLG_DO: {
               var
                  id = rq.decodeUint(), // Dialog id.
                  dlg = self.dlg.get(id),
                  d = {
                     f : rq.decodeString() // Focus object id
                  };
               if (dlg === undefined)
                  throw new Error("webdReceived: DLG " + id +
                     " does not exist, mode: " + rq.mode);
               if (!dlg.tabOrderValid) {
                  dlg.updateTabOrder();
                  d.t = dlg.tabOrder;
               }
               self.currDlg = dlg;
               self.pendingCall = true;
               rq.msg.push({
                  o : eq.MainOp.dlg_DO,
                  i : id,
                  d : d
               });
               break;
            }

            case md.POPUP_BOX: {
               if (self.resumeSession) {
                  self.resumeSession = undefined;
                  self.invalidateDlg(rq);
                  rq.msg.push({ o : eq.MainOp.resume });
               }

               var popup = { btn : [] }, btn;

               popup.x = rq.decodeInt();
               popup.y = rq.decodeInt();
               popup.title = rq.decodeString();
               popup.msg = rq.decodeString();
               popup.btnDefault = rq.decodeInt();
               while (btn = rq.decodeString())
                  popup.btn.push(btn);

               self.pendingCall = true;
               rq.msg.push({
                  o : eq.MainOp.dlg_POPUP_BOX,
                  d : popup
               });
               break;
            }

            default:
               throw new Error("webdReceived: invalid mode: " + rq.mode);
         }
      }

      if (rq.msg.length) {
         // Notify main thread.
         postMessage(rq.msg);
      }

      rq.dispose();
   }
   catch (e) {
      rq.dispose();
      throw e;
   }
};

/*------------------------------------------------------------------------
   static Api.webdOpened()
   WebSocket connection to webd server opened.
------------------------------------------------------------------------*/

eq.c.Api.webdOpened = function(e) {
   var self = eq.api;
   if (self.webdHandshakeTimer !== undefined)
      throw new Error( "BUG: webdOpened pending handshake");
   self.webdOpened = true;
   eq.c.Api.webdStopRetry();
   self.webdHandshakeTimer = setTimeout(eq.c.Api.webdHandshakeFailed, 30000);
};

eq.c.Api.webdHandshakeFailed = function() {
   console.log("WebSocket handshake failed");
   eq.c.Api.webdStopHandshakeTimer();
   eq.api.ws.close(1002); // Protocol error.
};

eq.c.Api.webdStopHandshakeTimer = function() {
   var self = eq.api;
   if (self.webdHandshakeTimer !== undefined) {
      clearTimeout(self.webdHandshakeTimer);
      self.webdHandshakeTimer = undefined;
   }
};

/*------------------------------------------------------------------------
   static Api.webdClosed()
   WebSocket connection to webd server closed.
------------------------------------------------------------------------*/

eq.c.Api.webdClosed = function(e) {
   var
      self = eq.api,
      api = eq.c.Api,
      retryTimer = self.webdRetryTimer,
      mayRetry = false,
      mayRestart = false,
      stopped = false;
   switch (e.code) {
      case 1000:
      case 1005: // IE
         mayRestart = true;
         // FALLTHROUGH
      case 1001:
         stopped = true;
         break;
      case 1006:
         mayRetry = true;
   }
   self.ws = undefined;
   if (retryTimer !== undefined && !mayRetry) {
      api.webdStopRetry();
      retryTimer = undefined;
      self.webdOpened = undefined;
   }
   self.startSession = undefined;
   self.resumeSession = undefined;
   api.webdStopHandshakeTimer();
   if (retryTimer === undefined) {
      self.webdOpened = undefined;
      if (stopped) {
         // Application has stopped.
         self.reset();
         self.resetDlg();
         // Notify main thread: Application has stopped.
         postMessage([{
            o : eq.MainOp.applicationStopped,
            d : mayRestart
         }]);
      }
   }
};

/*------------------------------------------------------------------------
   static Api.webdFailed()
   WebSocket connection to webd server closed.
------------------------------------------------------------------------*/

eq.c.Api.webdFailed = function(e) {
   var self = eq.api;
   if (   self.webdRetryTimer === undefined
       && (self.webdRetryCnt = self.reconnectTimeout) > 0) {
      self.webdRetryTimer = setInterval(eq.c.Api.webdRetry, 1000);
      console.log("WebSocket connection failed, retry: " +
         self.webdRetryCnt + " seconds");
      // Notify main thread: Cancel current call.
      postMessage([{ o : eq.MainOp.cancelCurrentCall }]);
   }
   else if (self.webdRetryCnt <= 0) {
      eq.c.Api.webdStopRetry();
      console.log("WebSocket connection failed");
      // Notify main thread: Application has failed.
      postMessage([{ o : eq.MainOp.applicationFailed }]);
   }
};

eq.c.Api.webdRetry = function() {
   var self = eq.api;
   if (self.webdRetryCnt > 0)
      if ((--self.webdRetryCnt % 10) === 0 && self.webdRetryCnt)
         console.log("WebSocket connection failed, retry: " +
            self.webdRetryCnt + " seconds");
   if (self.ws === undefined)
      self.webdOpen();
};

eq.c.Api.webdStopRetry = function() {
   var self = eq.api;
   if (self.webdRetryTimer !== undefined) {
      clearInterval(self.webdRetryTimer);
      self.webdRetryTimer = undefined;
      self.webdRetryCnt = undefined;
   }
};

/*========================================================================
   Start API thread.
========================================================================*/

eq.api = new eq.c.Api();
self.onmessage = eq.c.Api.mainReceived;

